import numpy as np
__author__ = 'David John Gagne <djgagne@ou.edu>'


def main():
    # Contingency Table from Wilks (2011) Table 8.3
    table = np.array([[50, 91, 71],
                      [47, 2364, 170],
                      [54, 205, 3288]])
    mct = MulticlassContingencyTable(table, n_classes=table.shape[0],
                                     class_names=np.arange(table.shape[0]).astype(str))
    print(mct.peirce_skill_score())
    print(mct.gerrity_score())


class MulticlassContingencyTable(object):
    """
    This class is a container for a contingency table containing more than 2 classes.
    The contingency table is stored in table as a numpy array with the rows corresponding to forecast categories,
    and the columns corresponding to observation categories.
    """
    def __init__(self, table=None, n_classes=2, class_names=("1", "0")):
        self.table = table
        self.n_classes = n_classes
        self.class_names = class_names
        if table is None:
            self.table = np.zeros((self.n_classes, self.n_classes), dtype=int)

    def __add__(self, other):
        assert self.n_classes == other.n_classes, "Number of classes does not match"
        return MulticlassContingencyTable(self.table + other.table,
                                          n_classes=self.n_classes,
                                          class_names=self.class_names)

    def peirce_skill_score(self):
        """
        Multiclass Peirce Skill Score (also Hanssen and Kuipers score, True Skill Score)
        """
        n = float(self.table.sum())
        nf = self.table.sum(axis=1)
        no = self.table.sum(axis=0)
        correct = float(self.table.trace())
        return (correct / n - (nf * no).sum() / n ** 2) / (1 - (no * no).sum() / n ** 2)

    def gerrity_score(self):
        """
        Gerrity Score, which weights each cell in the contingency table by its observed relative frequency.
        :return:
        """
        k = self.table.shape[0]
        n = float(self.table.sum())
        p_o = self.table.sum(axis=0) / n
        p_sum = np.cumsum(p_o)[:-1]
        a = (1.0 - p_sum) / p_sum
        s = np.zeros(self.table.shape, dtype=float)
        for (i, j) in np.ndindex(*s.shape):
            if i == j:
                s[i, j] = 1.0 / (k - 1.0) * (np.sum(1.0 / a[0:j]) + np.sum(a[j:k-1]))
            elif i < j:
                s[i, j] = 1.0 / (k - 1.0) * (np.sum(1.0 / a[0:i]) - (j - i) + np.sum(a[j:k-1]))
            else:
                s[i, j] = s[j, i]
        return np.sum(self.table / float(self.table.sum()) * s)

    def heidke_skill_score(self):
        n = float(self.table.sum())
        nf = self.table.sum(axis=1)
        no = self.table.sum(axis=0)
        correct = float(self.table.trace())
        return (correct / n - (nf * no).sum() / n ** 2) / (1 - (nf * no).sum() / n ** 2)


if __name__ == "__main__":
    main()



import pandas as pd
import numpy as np
import json
import argparse
from os.path import exists
from multiprocessing import Pool
from glob import glob
from datetime import datetime
import traceback

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--csv", help="CSV data file directory")
    parser.add_argument("-j", "--json", help="JSON forecast file directory")
    parser.add_argument("-o", "--out", help="Output path")
    parser.add_argument("-s", "--start", help="Start run date in YYYYMMDD format")
    parser.add_argument("-e", "--end", help="End run date in YYYYMMDD format")
    parser.add_argument("-c", "--cond", help="Condition model list (comma separated)")
    parser.add_argument("-m", "--dist", help="Size distribution model list (comma separated)")
    parser.add_argument("-n", "--ens", default="SSEF", help="Ensemble system name.")
    parser.add_argument("-p", "--proc", type=int, help="Number of processors")
    args = parser.parse_args()
    start_date = datetime.strptime(args.start, "%Y%m%d")
    end_date = datetime.strptime(args.end, "%Y%m%d")
    condition_models = args.cond.split(",")
    dist_models = args.dist.split(",")
    for condition_model in condition_models:
        print(condition_model)
    for dist_model in dist_models:
        print(dist_model)
    pool = Pool(args.proc)

    def output_combined_files(output):
        out_file = args.out + "track_forecast_data_{0}_{1}_combined.csv".format(*output[1:])
        if exists(out_file):
            output[0].to_csv(out_file, mode="a", header=False)
        else:
            output[0].to_csv(out_file, mode="w", index_label="Step_ID")
        return
    csv_files = sorted(glob(args.csv + "track_step_{0}_*.csv".format(args.ens)))
    print(csv_files)
    for csv_file in csv_files:
        run_date = datetime.strptime(csv_file[:-4].split("_")[-1], "%Y%m%d")
        if start_date <= run_date <= end_date:
            pool.apply_async(merge_input_csv_forecast_json,
                             (csv_file, args.json, condition_models, dist_models),
                             callback=output_combined_files)
    pool.close()
    pool.join()
    return


def merge_input_csv_forecast_json(input_csv_file, forecast_json_path, condition_models, dist_models):
    """
    Reads forecasts from json files and merges them with the input data from the step csv files.

    Args:
        input_csv_file: Name of the input data csv file being processed
        forecast_json_path: Path to the forecast json files toplevel directory
        condition_models: List of models used to forecast hail or no hail
        dist_models: List of models used to forecast the hail size distribution

    Returns:

    """
    try:
        run_date = input_csv_file[:-4].split("_")[-1]
        print(run_date)
        ens_member = "_".join(input_csv_file.split("/")[-1][:-4].split("_")[3:-1])
        ens_name = input_csv_file.split("/")[-1].split("_")[2]
        input_data = pd.read_csv(input_csv_file, index_col="Step_ID")
        full_json_path = forecast_json_path + "{0}/{1}/".format(run_date, ens_member)
        track_ids = sorted(input_data["Track_ID"].unique())
        model_pred_cols = []
        condition_models_ns = []
        dist_models_ns = []
        gamma_params = ["Shape", "Location", "Scale"]
        for condition_model in condition_models:
            model_pred_cols.append(condition_model.replace(" ", "-") + "_Condition")
            condition_models_ns.append(condition_model.replace(" ", "-"))
        for dist_model in dist_models:
            dist_models_ns.append(dist_model.replace(" ", "-"))
            for param in gamma_params:
                model_pred_cols.append(dist_model.replace(" ", "-") + "_" + param)
        pred_data = pd.DataFrame(index=input_data.index, columns=model_pred_cols,
                                dtype=float)
        for track_id in track_ids:
            track_id_num = track_id.split("_")[-1]
            json_filename = full_json_path + "{0}_{1}_{2}_model_track_{3}.json".format(ens_name,
                                                                                    run_date,
                                                                                    ens_member,
                                                                                    track_id_num)
            json_file = open(json_filename)
            json_data = json.load(json_file)
            json_file.close()
            for s, step in enumerate(json_data["features"]):
                step_id = track_id + "_{0:02d}".format(s)
                for cond_model in condition_models_ns:
                    pred_data.loc[step_id, cond_model + "_Condition"]  = step["properties"]["condition_" + cond_model]
                for dist_model in dist_models_ns:
                    pred_data.loc[step_id, [dist_model + "_" + p
                                            for p in gamma_params]] = step["properties"]["dist_" + dist_model]
        out_data = input_data.merge(pred_data, left_index=True, right_index=True)
        return out_data, ens_name, ens_member
    except Exception as e:
        print(traceback.format_exc())
        raise e
if __name__ == "__main__":
    main()




import numpy as np
import pandas as pd
from pyproj import Proj
import pygrib
from scipy.spatial import cKDTree
from scipy.ndimage import gaussian_filter
from os.path import exists


class HailForecastGrid(object):
    """
    HailForecastGrid loads and stores gridded machine learning hail forecasts from GRIB2 files. It can load
    an arbitrary number of members and timesteps at once.

    Attributes:
        run_date (datetime.datetime): Date of the initial time of the model run
        start_date (datetime.datetime): Date of the initial forecast time being loaded
        end_date (datetime.datetime): Date of the final forecast time being loaded
        forecast_dates (pandas.DatetimeIndex): All forecast times
        ensemble_name (str): Name of the NWP ensemble being used
        ml_model (str): Name of the machine learning model being loaded
        variable (str): Name of the machine learning model variable being forecast
        message_number (int): Field in the GRIB2 file to load. The first field in the file has message number 1.
        path (str): Path to top-level GRIB2 directory. Assumes files are stored in directories by run_date
        data (ndarray): Hail forecast data with dimensions (member, time, y, x)
        lon (ndarray): 2D array of longitudes
        lat (ndarray): 2D array of latitudes
        x (ndarray): 2d array of x-coordinate values in km
        y (ndarray): 2d array of y-coordinate values in km
        i (ndarray): 2d array of row indices
        j (ndarray): 2d array of column indices
        dx (float): distance between grid points
        proj (Proj): a pyproj projection object used for converting lat-lon points to x-y coordinate values
        projparams (dict): PROJ4 parameters describing map projection
    """
    def __init__(self, run_date, start_date, end_date, ensemble_name, ml_model, members,
                 variable, message_number, path):
        self.run_date = run_date
        self.start_date = start_date
        self.end_date = end_date
        self.forecast_dates = pd.DatetimeIndex(start=self.start_date, end=self.end_date, freq="1H")
        self.ensemble_name = ensemble_name
        self.ml_model = ml_model
        self.members = members
        self.variable = variable
        self.message_number = message_number
        self.path = path
        self.data = None
        self.lon = None
        self.lat = None
        self.x = None
        self.y = None
        self.i = None
        self.j = None
        self.dx = None
        self.proj = None
        self.projparams = None
        return

    def load_data(self):
        for m, member in enumerate(self.members):
            for f, forecast_date in enumerate(self.forecast_dates.to_pydatetime()):
                dt = int((forecast_date - self.run_date).total_seconds() / 3600)
                filename_args = (self.ensemble_name, member, self.ml_model, self.variable,
                                 forecast_date.strftime("%Y%m%d%H%M"))
                filename = self.path + self.run_date.strftime("%Y%m%d") + \
                           "/{0}_{1}_{2}_{3}_{4}.grib2".format(*filename_args)
                if not exists(filename):
                    filename_args = (self.ensemble_name, member, self.ml_model, self.variable,
                                     self.run_date.strftime("%Y%m%d%H") + "f{0:02d}".format(dt))
                    filename = self.path + self.run_date.strftime("%Y%m%d") + \
                               "/{0}_{1}_{2}_{3}_{4}.grib2".format(*filename_args)
                    if not exists(filename):
                        continue
                grbs = pygrib.open(filename)
                if self.lon is None:
                    self.lat, self.lon = grbs[self.message_number].latlons()
                    self.projparams = grbs[self.message_number].projparams
                    self.proj = Proj(grbs[self.message_number].projparams)
                    self.x, self.y = self.proj(self.lon, self.lat)
                    self.x /= 1000.0
                    self.y /= 1000.0
                    self.dx = grbs[self.message_number]['DxInMetres'] / 1000.0
                    self.i, self.j = np.indices(self.lon.shape)
                data = grbs[self.message_number].values
                data *= 1000.0
                if self.data is None:
                    self.data = np.empty((len(self.members), len(self.forecast_dates),
                                          data.shape[0], data.shape[1]), dtype=float)
                self.data[m, f] = data.filled(0)
                grbs.close()
        return

    def period_neighborhood_probability(self, radius, smoothing, threshold, stride,start_time,end_time):
        """
        Calculate the neighborhood probability over the full period of the forecast

        Args:
            radius: circular radius from each point in km
            smoothing: width of Gaussian smoother in km
            threshold: intensity of exceedance
            stride: number of grid points to skip for reduced neighborhood grid

        Returns:
            (neighborhood probabilities)
        """
        neighbor_x = self.x[::stride, ::stride]
        neighbor_y = self.y[::stride, ::stride]
        neighbor_kd_tree = cKDTree(np.vstack((neighbor_x.ravel(), neighbor_y.ravel())).T)
        neighbor_prob = np.zeros((self.data.shape[0], neighbor_x.shape[0], neighbor_x.shape[1]))
        print('Forecast Hours: {0}-{1}'.format(start_time, end_time))
        for m in range(len(self.members)):
            period_max = self.data[m,start_time:end_time,:,:].max(axis=0)
            valid_i, valid_j = np.where(period_max >= threshold)
            print(self.members[m], len(valid_i))
            if len(valid_i) > 0:
                var_kd_tree = cKDTree(np.vstack((self.x[valid_i, valid_j], self.y[valid_i, valid_j])).T)
                exceed_points = np.unique(np.concatenate(var_kd_tree.query_ball_tree(neighbor_kd_tree, radius))).astype(int)
                exceed_i, exceed_j = np.unravel_index(exceed_points, neighbor_x.shape)
                neighbor_prob[m][exceed_i, exceed_j] = 1
                if smoothing > 0:
                    neighbor_prob[m] = gaussian_filter(neighbor_prob[m], smoothing,mode='constant')
        return neighbor_prob



from .ModelGrid import ModelGrid
import numpy as np
import os
from os.path import exists


class SSEFModelGrid(ModelGrid):
    """
    Extension of ModelGrid to the CAPS Storm-Scale Ensemble Forecast system.

    Args:
        member (str): Name of the ensemble member
        run_date (datetime.datetime object): Date of the initial step of the ensemble run
        start_date (datetime.datetime object): First time step extracted.
        end_date (datetime.datetime object): Last time step extracted.
        path (str): Path to model output files.
        single_step (boolean (default=False)): Whether variable information is stored with each time step in a separate
            file or one file containing all timesteps.
    """
    def __init__(self, member, run_date, variable, start_date, end_date, path, single_step=False):
        self.path = path
        self.member = member
        forecast_hours = np.arange((start_date - run_date).total_seconds() / 3600,
                                   (end_date - run_date).total_seconds() / 3600 + 1)
        if single_step:
            full_path = self.path + "/".join([member, run_date.strftime("%Y%m%d"), "0000Z", "data2d"]) + "/"
            if not exists(full_path):
                full_path = self.path + "/".join([run_date.strftime("%Y%m%d"), member, "0000Z", "data2d"]) + "/"
        else:
            full_path = self.path + "/".join([member, run_date.strftime("%Y%m%d")]) + "/"
        potential_filenames = []
        if single_step:
            for hour in forecast_hours:
                potential_filenames.append("{0}ar{1}00.net{2}{3:06d}".format(full_path, 
                                                                             run_date.strftime("%Y%m%d"),
                                                                             variable.ljust(6,"_"),
                                                                             int(hour) * 3600))
        else:
            potential_filenames.append("{0}ssef_{1}_{2}_{3}.nc".format(full_path,
                                                                       self.member,
                                                                       run_date.strftime("%Y%m%d"),
                                                                       variable))
        filenames = []
        for filename in potential_filenames:
            if os.access(filename, os.R_OK):
                filenames.append(filename)
        super(SSEFModelGrid, self).__init__(filenames, run_date, start_date, end_date, variable)
        return



#!/usr/bin/env python
"""
Read a scikit-learn tree ensemble object and output the object into a human-readable text format.
"""
import pickle
import argparse


__author__ = "David John Gagne <djgagne@ou.edu>"
__copyright__ = "Copyright 2015, David John Gagne"
__email__ = "djgagne@ou.edu"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", required=True, help="Input scikit-learn tree ensemble model pickle file.")
    parser.add_argument("-o", "--output", required=True, help="Output file start.")
    parser.add_argument("-a", "--attr", default="", help="Attribute list file.")
    args = parser.parse_args()
    tree_ensemble_obj = load_tree_object(args.input)
    if args.attr != "":
        with open(args.attr) as attr_obj:
            attribute_names = attr_obj.readlines()
    else:
        attribute_names = None
    output_tree_ensemble(tree_ensemble_obj, args.output, attribute_names)
    return


def load_tree_object(filename):
    """
    Load scikit-learn decision tree ensemble object from file.
    
    Parameters
    ----------
    filename : str
        Name of the pickle file containing the tree object.
    
    Returns
    -------
    tree ensemble object
    """
    with open(filename) as file_obj:
        tree_ensemble_obj = pickle.load(file_obj)
    return tree_ensemble_obj


def output_tree_ensemble(tree_ensemble_obj, output_filename, attribute_names=None):
    """
    Write each decision tree in an ensemble to a file.

    Parameters
    ----------
    tree_ensemble_obj : sklearn.ensemble object
        Random Forest or Gradient Boosted Regression object
    output_filename : str
        File where trees are written
    attribute_names : list
        List of attribute names to be used in place of indices if available.
    """
    for t, tree in enumerate(tree_ensemble_obj.estimators_):
        print("Writing Tree {0:d}".format(t))
        out_file = open(output_filename + ".{0:d}.tree", "w")
        #out_file.write("Tree {0:d}\n".format(t))
        tree_str = print_tree_recursive(tree.tree_, 0, attribute_names)
        out_file.write(tree_str)
        #out_file.write("\n")
        out_file.close()
    return


def print_tree_recursive(tree_obj, node_index, attribute_names=None):
    """
    Recursively writes a string representation of a decision tree object.

    Parameters
    ----------
    tree_obj : sklearn.tree._tree.Tree object
        A base decision tree object
    node_index : int
        Index of the node being printed
    attribute_names : list
        List of attribute names
    
    Returns
    -------
    tree_str : str
        String representation of decision tree in the same format as the parf library.
    """
    tree_str = ""
    if node_index == 0:
        tree_str += "{0:d}\n".format(tree_obj.node_count)
    if tree_obj.feature[node_index] >= 0:
        if attribute_names is None:
            attr_val = "{0:d}".format(tree_obj.feature[node_index])
        else:
            attr_val = attribute_names[tree_obj.feature[node_index]]
        tree_str += "b {0:d} {1} {2:0.4f} {3:d} {4:1.5e}\n".format(node_index,
                                                                   attr_val,
                                                                   tree_obj.weighted_n_node_samples[node_index],
                                                                   tree_obj.n_node_samples[node_index],
                                                                   tree_obj.threshold[node_index])
    else:
        if tree_obj.max_n_classes > 1:
            leaf_value = "{0:d}".format(tree_obj.value[node_index].argmax())
        else:
            leaf_value = "{0}".format(tree_obj.value[node_index][0][0])
        tree_str += "l {0:d} {1} {2:0.4f} {3:d}\n".format(node_index,
                                                          leaf_value,
                                                          tree_obj.weighted_n_node_samples[node_index],
                                                          tree_obj.n_node_samples[node_index])
    if tree_obj.children_left[node_index] > 0:
        tree_str += print_tree_recursive(tree_obj, tree_obj.children_left[node_index], attribute_names)
    if tree_obj.children_right[node_index] > 0:
        tree_str += print_tree_recursive(tree_obj, tree_obj.children_right[node_index], attribute_names)
    return tree_str

if __name__ == "__main__":
    main()



from .ModelGrid import ModelGrid
from glob import glob
import pandas as pd
from datetime import timedelta
import numpy as np
import os, pdb
from os.path import exists

class VSEModelGrid(ModelGrid):
    """
    Extension of ModelGrid to VSE

    Args:
        member (str): Name of the ensemble member
        run_date (datetime.datetime object): Date of the initial step of the ensemble run
        start_date (datetime.datetime object): First time step extracted.
        end_date (datetime.datetime object): Last time step extracted.
        path (str): Path to model output files.
        single_step (boolean (default=False)): Whether variable information is stored with each time step in a separate
            file or one file containing all timesteps.
    """
    def __init__(self, member, run_date, variable, start_date, end_date, path, single_step=False):
        self.path = path
        self.member = member
        forecast_hours = np.arange((start_date - run_date).total_seconds() / 3600,
                                   (end_date - run_date).total_seconds() / 3600 + 1)
        # Maybe you need the vse-style file.
        if variable in ["XLAT","XLONG", "MU", "MUB", "Q2", "T2", "PSFC", "U10", "V10", "HGT", "RAINNC", "GRAUPELNC", "HAILNC", "PBLH", "WSPD10MAX", "W_UP_MAX", "W_DN_MAX", "REFD_MAX", "UP_HELI_MAX", "UP_HELI_MAX03", "UP_HELI_MAX01", "UP_HELI_MIN", "RVORT1_MAX", "RVORT0_MAX", "W_MEAN", "GRPL_MAX", "C_PBLH", "W_PBLH", "W_MAX_PBL", "W_1KM_AGL", "HAIL_MAXK1", "HAIL_MAX2D", "PREC_ACC_NC", "REFD_COM", "REFD", "ECHOTOP", "AFWA_MSLP", "AFWA_HAIL", "AFWA_HAIL_NEWMEAN", "AFWA_HAIL_NEWSTD"]:
                full_path = "/".join([self.path, member, run_date.strftime("%Y%m%d%H"), "wrf"])
                potential_filenames = []
                for hour in forecast_hours:
                    valid_time = run_date + timedelta(hours=hour)
                    potential_filenames.append(full_path +"/vse_d01."+valid_time.strftime("%Y-%m-%d_%H:%M:%S")+".nc")

        else: # use the WRFTWO??.nc file
                full_path = "/".join([self.path, member, run_date.strftime("%Y%m%d%H"), "post_AGAIN"])
                potential_filenames = []
                for hour in forecast_hours:
                    potential_filenames.append("{0}/fhr_{1:d}/WRFTWO{2:02d}.nc".format(full_path, int(hour), int(hour) ))
        filenames = []
        for filename in potential_filenames:
            if os.access(filename, os.R_OK):
                filenames.append(filename)
            else:
                print(filename, "not readable. dropping from list")
        super(VSEModelGrid, self).__init__(filenames, run_date, start_date, end_date, variable)
        return



import pandas as pd
import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", help="input csv file")
    parser.add_argument("-o", "--out", help="Output grib table file")
    args = parser.parse_args()
    grib_csv_table = pd.read_csv(args.input, skiprows=5)
    grib_csv_table.loc[:, "mtab_set"] = 1
    grib_csv_table.loc[:, "mtab_low"] = 0
    grib_csv_table.loc[:, "mtab_high"] = 10
    grib_csv_table.loc[:, "CenterNumber"] = 59
    grib_csv_table.loc[:, "local_version"] = 0
    # Order info from http://www.cpc.ncep.noaa.gov/products/wesley/wgrib2/user_grib2tables.html
    out_order = ["DisciplineNumber", "mtab_set", "mtab_low", "mtab_high", "CenterNumber", "local_version",
                 "CategoryNumber", "ParameterNumber", "NCLName", "Description", "Units"]
    grib_out_table = grib_csv_table[out_order]
    print(grib_out_table)
    grib_out_table.to_csv(args.out, sep=":", header=None, index=False)
    return


if __name__ == "__main__":
    main()


#
# The Enhanced Watershed code is based on the algorithm from Lakshmanan et al. (2009) and was adapted
# from Java code found at https://github.com/lakshmanok/asgbook/blob/master/src/edu/ou/asgbook/segmentation/EnhancedWatershedSegmenter.java.
#
# References
# Valliappa Lakshmanan, Kurt Hondl, and Robert Rabin, 2009: An Efficient, General-Purpose 
#  Technique for Identifying Storm Cells in Geospatial Images. J. Atmos. Oceanic Technol., 26, 523-537.
"""
@author: David John Gagne (djgagne@ou.edu)
"""

import numpy as np
from scipy.ndimage import label as splabel
from scipy.ndimage import find_objects
from collections import OrderedDict


class EnhancedWatershed(object):
    """
    The enhanced watershed performs image segmentation using a modified version of the traditional watershed technique.
    It includes a size criteria and creates foothills around each object to keep them distinct. The object is used to store
    the quantization and size parameters. It can be used to watershed multiple grids.

    Attributes:
        min_thresh (int): minimum pixel value for pixel to be part of a region
        data_increment (int): quantization interval. Use 1 if you don't want to quantize
        max_thresh (int): values greater than maxThresh are treated as the maximum threshold
        size_threshold_pixels (int): clusters smaller than this threshold are ignored.
        delta (int): maximum number of data increments the cluster is allowed to range over. Larger d results in clusters over larger scales.
    """
    def __init__(self, min_thresh, data_increment, max_thresh, size_threshold_pixels, delta):
        self.min_thresh = min_thresh
        self.data_increment = data_increment
        self.max_thresh = max_thresh
        self.max_size = size_threshold_pixels
        self.delta = delta
        self.max_bin = int((self.max_thresh - self.min_thresh) / self.data_increment)
        self.UNMARKED = -1
        self.GLOBBED = -3
        self.TOOSMALL = -4

    def label(self, input_grid):
        """
        Labels input grid using enhanced watershed algorithm.

        Args:
            input_grid (numpy.ndarray): Grid to be labeled.

        Returns:
            Array of labeled pixels
        """
        marked = self.find_local_maxima(input_grid)
        marked = np.where(marked >= 0, 1, 0)
        # splabel returns two things in a tuple: an array and an integer
        # assign the first thing (array) to markers
        markers = splabel(marked)[0]
        return markers

    @staticmethod
    def size_filter(labeled_grid, min_size):
        """
        Removes labeled objects that are smaller than minSize, and relabels the remaining objects.

        Args:
            labeled_grid: Grid that has been labeled
            min_size: Minimium object size.
        Returns:
            Labeled array with re-numbered objects to account for those that have been removed
        """
        out_grid = np.zeros(labeled_grid.shape,dtype=int)
        slices = find_objects(labeled_grid)
        j = 1
        for i, s in enumerate(slices):
            box = labeled_grid[s]
            size = np.count_nonzero(box == i + 1)
            if size >= min_size and box.shape[0] > 1 and box.shape[1] > 1:
                out_grid[np.where(labeled_grid == i+1)] = j
                j += 1
        return out_grid

    def find_local_maxima(self, input_grid):
        """
        Finds the local maxima in the inputGrid and perform region growing to identify objects.

        Args:
            input_grid: Raw input data.

        Returns:
            array with labeled objects.
        """
        pixels, q_data = self.quantize(input_grid)
        centers = OrderedDict()
        for p in pixels.keys():
            centers[p] = []
        marked = np.ones(q_data.shape, dtype=int) * self.UNMARKED
        MIN_INFL = int(np.round(1 + 0.5 * np.sqrt(self.max_size)))
        MAX_INFL = 2 * MIN_INFL
        marked_so_far = []
        # Find the maxima. These are high-values with enough clearance
        # around them.
        # Work from high to low bins. The pixels in the highest bin mark their
        # neighborhoods first. If you did it from low to high the lowest maxima
        # would mark their neighborhoods first and interfere with the identification of higher maxima.
        for b in sorted(pixels.keys(),reverse=True):
            # Square starts large with high intensity bins and gets smaller with low intensity bins.
            infl_dist = MIN_INFL + int(np.round(float(b) / self.max_bin * (MAX_INFL - MIN_INFL)))
            for p in pixels[b]:
                if marked[p] == self.UNMARKED:
                    ok = False
                    del marked_so_far[:]
                    # Temporarily mark unmarked points in square around point (keep track of them in list marked_so_far).
                    # If none of the points in square were marked already from a higher intensity center, 
                    # this counts as a new center and ok=True and points will remain marked.
                    # Otherwise ok=False and marked points that were previously unmarked will be unmarked.
                    for (i, j), v in np.ndenumerate(marked[p[0] - infl_dist:p[0] + infl_dist + 1,
                                                    p[1] - infl_dist:p[1]+ infl_dist + 1]):
                        if v == self.UNMARKED:
                            ok = True
                            marked[i - infl_dist + p[0],j - infl_dist + p[1]] = b
                           
                            marked_so_far.append((i - infl_dist + p[0],j - infl_dist + p[1]))
                        else:
                            # neighborhood already taken
                            ok = False
                            break
                    # ok if point and surrounding square were not marked already.
                    if ok:
                        # highest point in its neighborhood
                        centers[b].append(p)
                    else:
                        for m in marked_so_far:
                            marked[m] = self.UNMARKED
        # Erase marks and start over. You have a list of centers now.
        marked[:, :] = self.UNMARKED
        deferred_from_last = []
        deferred_to_next = []
        # delta (int): maximum number of increments the cluster is allowed to range over. Larger d results in clusters over larger scales.
        for delta in range(0, self.delta + 1):
            # Work from high to low bins.
            for b in sorted(centers.keys(), reverse=True):
                bin_lower = b - delta
                deferred_from_last[:] = deferred_to_next[:]
                del deferred_to_next[:]
                foothills = []
                n_centers = len(centers[b])
                tot_centers = n_centers + len(deferred_from_last)
                for i in range(tot_centers):
                    # done this way to minimize memory overhead of maintaining two lists
                    if i < n_centers:
                        center = centers[b][i]
                    else:
                        center = deferred_from_last[i - n_centers]
                    if bin_lower < 0:
                        bin_lower = 0
                    if marked[center] == self.UNMARKED:
                        captured = self.set_maximum(q_data, marked, center, bin_lower, foothills)
                        if not captured:
                            # decrement to lower value to see if it'll get big enough
                            deferred_to_next.append(center)
                        else:
                            pass
                # this is the last one for this bin
                self.remove_foothills(q_data, marked, b, bin_lower, centers, foothills)
            del deferred_from_last[:]
            del deferred_to_next[:]
        return marked
            
    def set_maximum(self, q_data, marked, center, bin_lower, foothills):
        """
        Grow a region at a certain bin level and check if the region has reached the maximum size.

        Args:
            q_data: Quantized data array
            marked: Array marking points that are objects
            center: Coordinates of the center pixel of the region being grown
            bin_lower: Intensity level of lower bin being evaluated
            foothills: List of points that are associated with a center but fall outside the the size or
                intensity criteria
        Returns:
            True if the object is finished growing and False if the object should be grown again at the next
            threshold level.
        """
        as_bin = [] # pixels to be included in peak
        as_glob = []   # pixels to be globbed up as part of foothills
        marked_so_far = []  # pixels that have already been marked
        will_be_considered_again = False
        as_bin.append(center)
        center_data = q_data[center]
        while len(as_bin) > 0:
            p = as_bin.pop(-1) # remove and return last pixel in as_bin
            if marked[p] != self.UNMARKED: # already processed
                continue
            marked[p] = q_data[center]
            marked_so_far.append(p)

            # check neighbors
            for index,val in np.ndenumerate(marked[p[0] - 1:p[0] + 2, p[1] - 1:p[1] + 2]):
                # is neighbor part of peak or part of mountain?
                if val == self.UNMARKED:
                    pixel = (index[0] - 1 + p[0],index[1] - 1 + p[1])
                    p_data = q_data[pixel]
                    if (not will_be_considered_again) and (p_data >= 0) and (p_data < center_data):
                        will_be_considered_again = True
                    if p_data >= bin_lower and (np.abs(center_data - p_data) <= self.delta):
                        as_bin.append(pixel)
                    # Do not check that this is the closest: this way, a narrow channel of globbed pixels form
                    elif p_data >= 0:
                        as_glob.append(pixel)
        if bin_lower == 0:
            will_be_considered_again = False
        big_enough = len(marked_so_far) >= self.max_size
        if big_enough:
            # remove lower values within region of influence
            foothills.append((center, as_glob))
        elif will_be_considered_again: # remove the check if you want to ignore regions smaller than max_size
            for m in marked_so_far:
                marked[m] = self.UNMARKED
            del as_bin[:]
            del as_glob[:]
            del marked_so_far[:]
        return big_enough or (not will_be_considered_again)

    def remove_foothills(self, q_data, marked, bin_num, bin_lower, centers, foothills):
        """
        Mark points determined to be foothills as globbed, so that they are not included in
        future searches. Also searches neighboring points to foothill points to determine
        if they should also be considered foothills.

        Args:
            q_data: Quantized data
            marked: Marked
            bin_num: Current bin being searched
            bin_lower: Next bin being searched
            centers: dictionary of local maxima considered to be object centers
            foothills: List of foothill points being removed.
        """
        hills = []
        for foot in foothills:
            center = foot[0]
            hills[:] = foot[1][:]
            # remove all foothills
            while len(hills) > 0:
                # mark this point
                pt = hills.pop(-1)
                marked[pt] = self.GLOBBED
                for s_index, val in np.ndenumerate(marked[pt[0]-1:pt[0]+2,pt[1]-1:pt[1]+2]):
                    index = (s_index[0] - 1 + pt[0], s_index[1] - 1 + pt[1])
                    # is neighbor part of peak or part of mountain?
                    if val == self.UNMARKED:
                        # will let in even minor peaks
                        if (q_data[index] >= 0) and \
                                (q_data[index] < bin_lower) and \
                                ((q_data[index] <= q_data[pt]) or self.is_closest(index, center, centers, bin_num)):
                            hills.append(index)
        del foothills[:] 

    @staticmethod
    def is_closest(point, center, centers, bin_num):
        bin_thresh = int(bin_num / 2)
        p_arr = np.array(point)
        c_arr = np.array(center)
        my_dist = np.sum(np.power(p_arr - c_arr, 2))
        for o_bin in range(bin_thresh, len(centers.keys())):
            for c in centers[o_bin]:
                oc_arr = np.array(c)
                if np.sum(np.power(p_arr - oc_arr, 2)) < my_dist:
                    return False
        return True

    def quantize(self, input_grid):
        """
        Quantize a grid into discrete steps based on input parameters.

        Args:
            input_grid: 2-d array of values

        Returns:
            Dictionary of value pointing to pixel locations, and quantized 2-d array of data
        """
        pixels = {}
        for i in range(self.max_bin+1):
            pixels[i] = []

        data = (np.array(input_grid, dtype=int) - self.min_thresh) / self.data_increment
        data[data < 0] = -1
        data[data > self.max_bin] = self.max_bin
        good_points = np.where(data >= 0)
        for g in np.arange(good_points[0].shape[0]):
            pixels[data[(good_points[0][g], good_points[1][g])]].append((good_points[0][g], good_points[1][g]))
        return pixels, data

    @staticmethod
    def is_valid(point, shape):
        return np.all((np.array(point) >= 0) & (np.array(shape) - np.array(point) > 0))


def rescale_data(data, data_min, data_max, out_min=0.0, out_max=100.0):
    """
    Rescale your input data so that is ranges over integer values, which will perform better in the watershed.

    Args:
        data: 2D or 3D ndarray being rescaled
        data_min: minimum value of input data for scaling purposes
        data_max: maximum value of input data for scaling purposes
        out_min: minimum value of scaled data
        out_max: maximum value of scaled data

    Returns:
        Linearly scaled ndarray
    """
    return (out_max - out_min) / (data_max - data_min) * (data - data_min) + out_min



import numpy as np
from create_sector_grid_data import *             
from datetime import datetime
           
from hagelslag.processing.ObjectMatcher import shifted_centroid_distance, start_time_distance
from hagelslag.processing.ObjectMatcher import centroid_distance, time_distance
import os
import pandas as pd
import numpy as np
from datetime import datetime


#date_index = pd.DatetimeIndex([pd.Timestamp.utcnow().strftime("%Y%m%d")])
work_path = "/ai-hail/aburke/2018_HREFv2_data/"
scratch_path= "/hail/aburke/testing_weights/figuring_out_files/"
ensemble_members = ['nam_00','nam_12','arw_00','arw_12','nssl_00','nssl_12','nmmb_00','nmmb_12']

config=dict(dates=None,
            start_hour=12, 
            end_hour=36,
            watershed_variable="MAXUVV",
            ensemble_name="HREFv2",
            ensemble_members=ensemble_members,
            model_path=work_path+'symbolic_hrefv2_data',
            model_watershed_params=(8, 1, 80, 100, 60),
            size_filter=12,
            gaussian_window=2,
            mrms_path=work_path+"MRMS/",
            mrms_variable="MESH_Max_60min_00.50",
            mrms_watershed_params=(19, 1, 100, 100, 75),
            object_matcher_params=([shifted_centroid_distance], np.array([1.0]),
                                     np.array([24000])),
        
            track_matcher_params=([centroid_distance, time_distance],
                                     np.array([80000, 2])),
		
            storm_variables=['MAXUVV','Storm relative helicity_3000', 'Storm relative helicity_1000',
                            'MAXREF','MXUPHL_5000','MAXDVV'], 

            potential_variables=['Precipitable water_0','Temperature_1000','Dew point temperature_1000','Geopotential Height_500','Temperature_500', 
                                'Dew point temperature_500','U component of wind_500','V component of wind_500',
                                'Geopotential Height_700','Temperature_700', 'Dew point temperature_700','U component of wind_700',
                                'V component of wind_700','Geopotential Height_850','Temperature_850', 'Dew point temperature_850',
                                'U component of wind_850','V component of wind_850','MAXUW', 'MAXVW',
                                'Surface lifted index','Convective available potential energy_0','Convective inhibition_0'],
        
            tendency_variables=[],
            shape_variables=["area", "eccentricity", "major_axis_length", "minor_axis_length", "orientation",
                               "extent"],
            variable_statistics=["mean", "max", "min", "std", "skew",
                                   "percentile_10", "percentile_50", "percentile_90"],
            csv_path=scratch_path+"track_data_spring2018_MAXUVV_closest_csv/",
            geojson_path=scratch_path+"track_data_spring2018_MAXUVV_closest_json/",
	        nc_path=scratch_path+"track_data_spring2018_MAXUVV_patch_nc/",
            unique_matches=True,
	        patch_radius=16,
            closest_matches=True,
	        match_steps=True,              
	        train=False,
            single_step=True,
            label_type="gamma",
            model_map_file="/hail/aburke/hagelslag/mapfiles/hrefv2_2018_map.txt",
            mask_file="/hail/aburke/hagelslag/mapfiles/hrefv2_us_mask.nc",
            run_date_format="%Y%m%d-%H%M",
            sector_csv_outpath=scratch_path+'track_data_spring2018_MAXUVV_closest_csv/',
            sector_nc_outpath=scratch_path+'track_data_spring2018_MAXUVV_patch_nc/',
            sector_mapfile="/hail/aburke/hagelslag/mapfiles/hrefv2_sectors/C_2018_map.txt"
            )

sector='E'

date_index = pd.DatetimeIndex(start="2018-05-01T00:00", end="2018-06-02T00:00", freq="1D")

sector_mapfile = "/hail/aburke/hagelslag/mapfiles/hrefv2_sectors/{0}_2018_map.txt".format(sector)
ensemble_name="HREFv2"
member='arw_00'
run_date = datetime(2018,7,1) 
run_date_format="%Y%m%d-%H%M" 
csv_path='/hail/aburke/HREF_Scripts_and_Data/hwt_2018/track_data_spring2018_MAXUVV_closest_csv/'
nc_path='/hail/aburke/HREF_Scripts_and_Data/hwt_2018/track_data_spring2018_MAXUVV_patch_nc/'
sector_csv_outpath="/hail/aburke/testing_weights/sector_trained/{0}/track_data_spring2018_MAXUVV_closest_csv/".format(sector)
sector_nc_outpath="/hail/aburke/testing_weights/sector_trained/{0}/track_data_spring2018_MAXUVV_patch_nc/".format(sector)
patch_radius=16

for run_date in date_index:
    for member in ensemble_members:
        sector = SectorProcessor(sector_mapfile,
                        ensemble_name,member,
                        run_date,run_date_format)

        sector.output_sector_netcdf(nc_path,sector_nc_outpath,patch_radius,config)

        #for key in ['track_step', 'track_total']:
        #    sector.output_sector_csv(csv_path,key,sector_csv_outpath)



from netCDF4 import Dataset
import numpy as np
from hagelslag.data.MRMSGrid import MRMSGrid
from hagelslag.evaluation.ProbabilityMetrics import DistributedROC, DistributedReliability
from datetime import timedelta
from scipy.ndimage import binary_dilation


class GridEvaluator(object):
    """
    An evaluation system for gridded forecasts.

    GridEvaluator loads in a set of machine learning model forecasts from a single model run, loads in corresponding
    observations, and then generates verification statistics from the matching of forecasts and observations. Forecasts
    can be aggregated in time with flexible window sizes.

    Parameters
    ----------
    run_date : datetime.datetime object
        The date of an ensemble run.
    ensemble_name : str
        Name of the ensemble. Should be consistent with the name used in the data processing and forecasting.
    ensemble_member : str
        Name of the ensemble member being loaded.
    model_names : list of strings
        Names of the machine learning models being evaluated.
    size_thresholds : list or numpy.ndarray of ints
        Intensity thresholds at which probability forecasts are made.
    start_hour : int
        Forecast hour at which evaluation begins.
    end_hour : int
        Forecast hour at which evaluation ends, inclusive.
    window_size : int
        Number of hours to include within a forecast window.
    time_skip : int
        Number of hours to skip between window starts
    forecast_path : str
        Path to where gridded forecasts are located.
    mrms_path : str
        Path to the MRMS gridded observations.
    mrms_variable: str
        Name of the variable being used for verification.
    obs_mask : bool, optional (default=True)
        Whether or not a masking grid is used to determine which grid points are evaluated
    mask_variable : str, optional (default="RadarQualityIndex_00.00")
        Name of the MRMS variable used for masking.
    """
    def __init__(self, run_date, ensemble_name, ensemble_member,
                 model_names, size_thresholds, start_hour, end_hour, window_size, time_skip,
                 forecast_path, mrms_path, mrms_variable, obs_mask=True,
                 mask_variable="RadarQualityIndex_00.00"):
        self.run_date = run_date
        self.ensemble_name = ensemble_name
        self.ensemble_member = ensemble_member
        self.model_names = model_names
        self.size_thresholds = size_thresholds
        self.start_hour = start_hour
        self.end_hour = end_hour
        self.valid_hours = np.arange(self.start_hour, self.end_hour + 1)
        self.window_size = window_size
        self.time_skip = time_skip
        self.hour_windows = []
        for hour in self.valid_hours[self.window_size-1::self.time_skip]:
            self.hour_windows.append(slice(hour-self.window_size+1 - self.start_hour,
                                           hour+1-self.start_hour))
        self.forecast_path = forecast_path
        self.mrms_path = mrms_path
        self.mrms_variable = mrms_variable
        self.raw_forecasts = {}
        self.raw_obs = {}
        self.window_forecasts = {}
        self.window_obs = {}
        self.dilated_obs = {}
        self.obs_mask = obs_mask
        self.mask_variable = mask_variable

    def load_forecasts(self):
        """
        Load the forecast files into memory.
        """
        run_date_str = self.run_date.strftime("%Y%m%d")
        for model_name in self.model_names:
            self.raw_forecasts[model_name] = {}
            forecast_file = self.forecast_path + run_date_str + "/" + \
                model_name.replace(" ", "-") + "_hailprobs_{0}_{1}.nc".format(self.ensemble_member, run_date_str)
            forecast_obj = Dataset(forecast_file)
            forecast_hours = forecast_obj.variables["forecast_hour"][:]
            valid_hour_indices = np.where((self.start_hour <= forecast_hours) & (forecast_hours <= self.end_hour))[0]
            for size_threshold in self.size_thresholds:
                self.raw_forecasts[model_name][size_threshold] = \
                    forecast_obj.variables["prob_hail_{0:02d}_mm".format(size_threshold)][valid_hour_indices]
            forecast_obj.close()

    def get_window_forecasts(self):
        """
        Aggregate the forecasts within the specified time windows.
        """
        for model_name in self.model_names:
            self.window_forecasts[model_name] = {}
            for size_threshold in self.size_thresholds:
                self.window_forecasts[model_name][size_threshold] = \
                    np.array([self.raw_forecasts[model_name][size_threshold][sl].sum(axis=0)
                              for sl in self.hour_windows])

    def load_obs(self,  mask_threshold=0.5):
        """
        Loads observations and masking grid (if needed).

        :param mask_threshold: Values greater than the threshold are kept, others are masked.
        :return:
        """
        start_date = self.run_date + timedelta(hours=self.start_hour)
        end_date = self.run_date + timedelta(hours=self.end_hour)
        mrms_grid = MRMSGrid(start_date, end_date, self.mrms_variable, self.mrms_path)
        mrms_grid.load_data()
        if len(mrms_grid.data) > 0:
            self.raw_obs[self.mrms_variable] = np.where(mrms_grid.data > 100, 100, mrms_grid.data)
            self.window_obs[self.mrms_variable] = np.array([self.raw_obs[self.mrms_variable][sl].max(axis=0)
                                                            for sl in self.hour_windows])
            if self.obs_mask:
                mask_grid = MRMSGrid(start_date, end_date, self.mask_variable, self.mrms_path)
                mask_grid.load_data()
                self.raw_obs[self.mask_variable] = np.where(mask_grid.data >= mask_threshold, 1, 0)
                self.window_obs[self.mask_variable] = np.array([self.raw_obs[self.mask_variable][sl].max(axis=0)
                                                               for sl in self.hour_windows])

    def dilate_obs(self, dilation_radius):
        """
        Use a dilation filter to grow positive observation areas by a specified number of grid points

        :param dilation_radius: Number of times to dilate the grid.
        :return:
        """
        for s in self.size_thresholds:
            self.dilated_obs[s] = np.zeros(self.window_obs[self.mrms_variable].shape)
            for t in range(self.dilated_obs[s].shape[0]):
                self.dilated_obs[s][t][binary_dilation(self.window_obs[self.mrms_variable][t] >= s, iterations=dilation_radius)] = 1

    def roc_curves(self, prob_thresholds):
        """
        Generate ROC Curve objects for each machine learning model, size threshold, and time window.

        :param prob_thresholds: Probability thresholds for the ROC Curve
        :param dilation_radius: Number of times to dilate the observation grid.
        :return: a dictionary of DistributedROC objects.
        """
        all_roc_curves = {}
        for model_name in self.model_names:
            all_roc_curves[model_name] = {}
            for size_threshold in self.size_thresholds:
                all_roc_curves[model_name][size_threshold] = {}
                for h, hour_window in enumerate(self.hour_windows):
                    hour_range = (hour_window.start, hour_window.stop)
                    all_roc_curves[model_name][size_threshold][hour_range] = \
                        DistributedROC(prob_thresholds, 1)
                    if self.obs_mask:
                        all_roc_curves[model_name][size_threshold][hour_range].update(
                            self.window_forecasts[model_name][size_threshold][h][
                                self.window_obs[self.mask_variable][h] > 0],
                            self.dilated_obs[size_threshold][h][self.window_obs[self.mask_variable][h] > 0]
                        )
                    else:
                        all_roc_curves[model_name][size_threshold][hour_range].update(
                            self.window_forecasts[model_name][size_threshold][h],
                            self.dilated_obs[size_threshold][h]
                        )
        return all_roc_curves

    def reliability_curves(self, prob_thresholds):
        """
        Output reliability curves for each machine learning model, size threshold, and time window.

        :param prob_thresholds:
        :param dilation_radius:
        :return:
        """
        all_rel_curves = {}
        for model_name in self.model_names:
            all_rel_curves[model_name] = {}
            for size_threshold in self.size_thresholds:
                all_rel_curves[model_name][size_threshold] = {}
                for h, hour_window in enumerate(self.hour_windows):
                    hour_range = (hour_window.start, hour_window.stop)
                    all_rel_curves[model_name][size_threshold][hour_range] = \
                        DistributedReliability(prob_thresholds, 1)
                    if self.obs_mask:
                        all_rel_curves[model_name][size_threshold][hour_range].update(
                            self.window_forecasts[model_name][size_threshold][h][
                                self.window_obs[self.mask_variable][h] > 0],
                            self.dilated_obs[size_threshold][h][self.window_obs[self.mask_variable][h] > 0]
                        )
                    else:
                        all_rel_curves[model_name][size_threshold][hour_range].update(
                            self.window_forecasts[model_name][size_threshold][h],
                            self.dilated_obs[size_threshold][h]
                        )
        return all_rel_curves











import numpy as np
from skimage.measure import regionprops
from skimage.segmentation import find_boundaries
from skimage.morphology import convex_hull_image
import json
import os


class STObject(object):
    """
    The STObject stores data and location information for objects extracted from the ensemble grids.

    Attributes:
        grid (ndarray): All of the data values. Supports a 2D array of values, a list of 2D arrays, or a 3D array.
        mask (ndarray): Grid of 1's and 0's in which 1's indicate the location of the object.
        x (ndarray): Array of x-coordinate values in meters. Longitudes can also be placed here.
        y (ndarray): Array of y-coordinate values in meters. Latitudes can also be placed here.
        i (ndarray): Array of row indices from the full model domain.
        j (ndarray): Array of column indices from the full model domain.
        start_time: The first time of the object existence.
        end_time: The last time of the object existence.
        step: number of hours between timesteps
        dx: grid spacing
        u: storm motion in x-direction
        v: storm motion in y-direction
    """

    def __init__(self, grid, mask, x, y, i, j, start_time, end_time, step=1, dx=4000, u=None, v=None):
        if hasattr(grid, "shape") and len(grid.shape) == 2:
            self.timesteps = [grid]
            self.masks = [np.array(mask, dtype=int)]
            self.x = [x]
            self.y = [y]
            self.i = [i]
            self.j = [j]
        elif hasattr(grid, "shape") and len(grid.shape) > 2:
            self.timesteps = []
            self.masks = []
            self.x = []
            self.y = []
            self.i = []
            self.j = []
            for l in range(grid.shape[0]):
                self.timesteps.append(grid[l])
                self.masks.append(np.array(mask[l], dtype=int))
                self.x.append(x[l])
                self.y.append(y[l])
                self.i.append(i[l])
                self.j.append(j[l])
        else:
            self.timesteps = grid
            self.masks = mask
            self.x = x
            self.y = y
            self.i = i
            self.j = j
        if u is not None and v is not None:
            self.u = u
            self.v = v
        else:
            self.u = np.zeros(len(self.timesteps))
            self.v = np.zeros(len(self.timesteps))
        self.dx = dx
        self.start_time = start_time
        self.end_time = end_time
        self.step = step
        self.times = np.arange(start_time, end_time + step, step)
        self.attributes = {}
        self.observations = None

    @property
    def __str__(self):
        com_x, com_y = self.center_of_mass(self.start_time)
        data = dict(maxsize=self.max_size(), comx=com_x, comy=com_y, start=self.start_time, end=self.end_time)
        return "ST Object [maxSize=%(maxsize)d,initialCenter=%(comx)0.2f,%(comy)0.2f,duration=%(start)02d-%(end)02d]" %\
               data

    def center_of_mass(self, time):
        """
        Calculate the center of mass at a given timestep.

        Args:
            time: Time at which the center of mass calculation is performed

        Returns:
            The x- and y-coordinates of the center of mass.
        """
        if self.start_time <= time <= self.end_time:
            diff = time - self.start_time
            valid = np.flatnonzero(self.masks[diff] != 0)
            if valid.size > 0:
                com_x = 1.0 / self.timesteps[diff].ravel()[valid].sum() * np.sum(self.timesteps[diff].ravel()[valid] *
                                                                                 self.x[diff].ravel()[valid])
                com_y = 1.0 / self.timesteps[diff].ravel()[valid].sum() * np.sum(self.timesteps[diff].ravel()[valid] *
                                                                                 self.y[diff].ravel()[valid])
            else:
                com_x = np.mean(self.x[diff])
                com_y = np.mean(self.y[diff])
        else:
            com_x = None
            com_y = None
        return com_x, com_y

    def closest_distance(self, time, other_object, other_time):
        """
        The shortest distance between two objects at specified times.

        Args:
            time (int or datetime): Valid time for this STObject
            other_object: Another STObject being compared
            other_time: The time within the other STObject being evaluated.

        Returns:
            Distance in units of the x-y coordinates
        """
        ti = np.where(self.times == time)[0][0]
        oti = np.where(other_object.times == other_time)[0][0]
        xs = self.x[ti].ravel()[self.masks[ti].ravel() == 1]
        xs = xs.reshape(xs.size, 1)
        ys = self.y[ti].ravel()[self.masks[ti].ravel() == 1]
        ys = ys.reshape(ys.size, 1)
        o_xs = other_object.x[oti].ravel()[other_object.masks[oti].ravel() == 1]
        o_xs = o_xs.reshape(1, o_xs.size)
        o_ys = other_object.y[oti].ravel()[other_object.masks[oti].ravel() == 1]
        o_ys = o_ys.reshape(1, o_ys.size)
        distances = (xs - o_xs) ** 2 + (ys - o_ys) ** 2
        return np.sqrt(distances.min())

    def percentile_distance(self, time, other_object, other_time, percentile):
        ti = np.where(self.times == time)[0][0]
        oti = np.where(other_object.times == other_time)[0][0]
        xs = self.x[ti][self.masks[ti] == 1]
        xs = xs.reshape(xs.size, 1)
        ys = self.y[ti][self.masks[ti] == 1]
        ys = ys.reshape(ys.size, 1)
        o_xs = other_object.x[oti][other_object.masks[oti] == 1]
        o_xs = o_xs.reshape(1, o_xs.size)
        o_ys = other_object.y[oti][other_object.masks[oti] == 1]
        o_ys = o_ys.reshape(1, o_ys.size)
        distances = (xs - o_xs) ** 2 + (ys - o_ys) ** 2
        return np.sqrt(np.percentile(distances, percentile))

    def trajectory(self):
        """
        Calculates the center of mass for each time step and outputs an array

        Returns:

        """
        traj = np.zeros((2, self.times.size))
        for t, time in enumerate(self.times):
            traj[:, t] = self.center_of_mass(time)
        return traj

    def get_corner(self, time):
        """
        Gets the corner array indices of the STObject at a given time that corresponds 
        to the upper left corner of the bounding box for the STObject.

        Args:
            time: time at which the corner is being extracted.

        Returns:
              corner index.
        """
        if self.start_time <= time <= self.end_time:
            diff = time - self.start_time
            return self.i[diff][0, 0], self.j[diff][0, 0]
        else:
            return -1, -1

    def size(self, time):
        """
        Gets the size of the object at a given time.

        Args:
            time: Time value being queried.

        Returns:
            size of the object in pixels
        """
        if self.start_time <= time <= self.end_time:
            return self.masks[time - self.start_time].sum()
        else:
            return 0

    def max_size(self):
        """
        Gets the largest size of the object over all timesteps.
        
        Returns:
            Maximum size of the object in pixels
        """
        sizes = np.array([m.sum() for m in self.masks])
        return sizes.max()

    def max_intensity(self, time):
        """
        Calculate the maximum intensity found at a timestep.

        """
        ti = np.where(time == self.times)[0][0]
        return self.timesteps[ti].max()

    def extend(self, step):
        """
        Adds the data from another STObject to this object.
        
        Args:
            step: another STObject being added after the current one in time.
        """
        self.timesteps.extend(step.timesteps)
        self.masks.extend(step.masks)
        self.x.extend(step.x)
        self.y.extend(step.y)
        self.i.extend(step.i)
        self.j.extend(step.j)
        self.end_time = step.end_time
        self.times = np.arange(self.start_time, self.end_time + self.step, self.step)
        self.u = np.concatenate((self.u, step.u))
        self.v = np.concatenate((self.v, step.v))
        for attr in self.attributes.keys():
            if attr in step.attributes.keys():
                self.attributes[attr].extend(step.attributes[attr])

    def boundary_polygon(self, time):
        """
        Get coordinates of object boundary in counter-clockwise order
        """
        ti = np.where(time == self.times)[0][0]
        com_x, com_y = self.center_of_mass(time)
        # If at least one point along perimeter of the mask rectangle is unmasked, find_boundaries() works.
        # But if all perimeter points are masked, find_boundaries() does not find the object.
        # Therefore, pad the mask with zeroes first and run find_boundaries on the padded array.
        padded_mask = np.pad(self.masks[ti], 1, 'constant', constant_values=0)
        chull = convex_hull_image(padded_mask)
        boundary_image = find_boundaries(chull, mode='inner', background=0)
        # Now remove the padding.
        boundary_image = boundary_image[1:-1,1:-1]
        boundary_x = self.x[ti].ravel()[boundary_image.ravel()]
        boundary_y = self.y[ti].ravel()[boundary_image.ravel()]
        r = np.sqrt((boundary_x - com_x) ** 2 + (boundary_y - com_y) ** 2)
        theta = np.arctan2((boundary_y - com_y), (boundary_x - com_x)) * 180.0 / np.pi + 360
        polar_coords = np.array([(r[x], theta[x]) for x in range(r.size)], dtype=[('r', 'f4'), ('theta', 'f4')])
        coord_order = np.argsort(polar_coords, order=['theta', 'r'])
        ordered_coords = np.vstack([boundary_x[coord_order], boundary_y[coord_order]])
        return ordered_coords

    def estimate_motion(self, time, intensity_grid, max_u, max_v):
        """
        Estimate the motion of the object with cross-correlation on the intensity values from the previous time step.

        Args:
            time: time being evaluated.
            intensity_grid: 2D array of intensities used in cross correlation.
            max_u: Maximum x-component of motion. Used to limit search area.
            max_v: Maximum y-component of motion. Used to limit search area

        Returns:
            u, v, and the minimum error.
        """
        ti = np.where(time == self.times)[0][0]
        mask_vals = np.where(self.masks[ti].ravel() == 1)
        i_vals = self.i[ti].ravel()[mask_vals]
        j_vals = self.j[ti].ravel()[mask_vals]
        obj_vals = self.timesteps[ti].ravel()[mask_vals]
        u_shifts = np.arange(-max_u, max_u + 1)
        v_shifts = np.arange(-max_v, max_v + 1)
        min_error = 99999999999.0
        best_u = 0
        best_v = 0
        for u in u_shifts:
            j_shift = j_vals - u
            for v in v_shifts:
                i_shift = i_vals - v
                if np.all((0 <= i_shift) & (i_shift < intensity_grid.shape[0]) &
                                  (0 <= j_shift) & (j_shift < intensity_grid.shape[1])):
                    shift_vals = intensity_grid[i_shift, j_shift]
                else:
                    shift_vals = np.zeros(i_shift.shape)
                # This isn't correlation; it is mean absolute error.
                error = np.abs(shift_vals - obj_vals).mean()
                if error < min_error:
                    min_error = error
                    best_u = u * self.dx
                    best_v = v * self.dx
        # 60 seems arbitrarily high
        #if min_error > 60:
        #    best_u = 0
        #    best_v = 0
        self.u[ti] = best_u
        self.v[ti] = best_v
        return best_u, best_v, min_error

    def count_overlap(self, time, other_object, other_time):
        """
        Counts the number of points that overlap between this STObject and another STObject. Used for tracking.
        """
        ti = np.where(time == self.times)[0][0]
        ma = np.where(self.masks[ti].ravel() == 1)
        oti = np.where(other_time == other_object.times)[0]
        obj_coords = np.zeros(self.masks[ti].sum(), dtype=[('x', int), ('y', int)])
        other_obj_coords = np.zeros(other_object.masks[oti].sum(), dtype=[('x', int), ('y', int)])
        obj_coords['x'] = self.i[ti].ravel()[ma]
        obj_coords['y'] = self.j[ti].ravel()[ma]
        other_obj_coords['x'] = other_object.i[oti][other_object.masks[oti] == 1]
        other_obj_coords['y'] = other_object.j[oti][other_object.masks[oti] == 1]
        return float(np.intersect1d(obj_coords,
                                    other_obj_coords).size) / np.maximum(self.masks[ti].sum(),
                                                                         other_object.masks[oti].sum())

    def extract_attribute_grid(self, model_grid, potential=False, future=False):
        """
        Extracts the data from a ModelOutput or ModelGrid object within the bounding box region of the STObject.
        
        Args:
            model_grid: A ModelGrid or ModelOutput Object
            potential: Extracts from the time before instead of the same time as the object
        """

        if potential:
            var_name = model_grid.variable + "-potential"
            timesteps = np.arange(self.start_time - 1, self.end_time)
        elif future:
            var_name = model_grid.variable + "-future"
            timesteps = np.arange(self.start_time + 1, self.end_time + 2)
        else:
            var_name = model_grid.variable
            timesteps = np.arange(self.start_time, self.end_time + 1)
        self.attributes[var_name] = []
        for ti, t in enumerate(timesteps):
            self.attributes[var_name].append(
                model_grid.data[t - model_grid.start_hour, self.i[ti], self.j[ti]])

    def extract_attribute_array(self, data_array, var_name):
        """
        Extracts data from a 2D array that has the same dimensions as the grid used to identify the object.

        Args:
            data_array: 2D numpy array

        """
        if var_name not in self.attributes.keys():
            self.attributes[var_name] = []
        for t in range(self.times.size):
            self.attributes[var_name].append(data_array[self.i[t], self.j[t]])


    def extract_tendency_grid(self, model_grid):
        """
        Extracts the difference in model outputs

        Args:
            model_grid: ModelOutput or ModelGrid object.

        """
        var_name = model_grid.variable + "-tendency"
        self.attributes[var_name] = []
        timesteps = np.arange(self.start_time, self.end_time + 1)
        for ti, t in enumerate(timesteps):
            t_index = t - model_grid.start_hour
            self.attributes[var_name].append(
                model_grid.data[t_index, self.i[ti], self.j[ti]] - model_grid.data[t_index - 1, self.i[ti], self.j[ti]]
                )

    def calc_attribute_statistics(self, statistic_name):
        """
        Calculates summary statistics over the domains of each attribute.
        
        Args:
            statistic_name (string): numpy statistic, such as mean, std, max, min

        Returns:
            dict of statistics from each attribute grid.
        """
        stats = {}
        for var, grids in self.attributes.items():
            if len(grids) > 1:
                stats[var] = getattr(np.array([getattr(np.ma.array(x, mask=self.masks[t] == 0), statistic_name)()
                                               for t, x in enumerate(grids)]), statistic_name)()
            else:
                stats[var] = getattr(np.ma.array(grids[0], mask=self.masks[0] == 0), statistic_name)()
        return stats

    def calc_attribute_statistic(self, attribute, statistic, time):
        """
        Calculate statistics based on the values of an attribute. The following statistics are supported:
        mean, max, min, std, ptp (range), median, skew (mean - median), and percentile_(percentile value).

        Args:
            attribute: Attribute extracted from model grid
            statistic: Name of statistic being used.
            time: timestep of the object being investigated

        Returns:
            The value of the statistic
        """
        ti = np.where(self.times == time)[0][0]
        ma = np.where(self.masks[ti].ravel() == 1)
        if statistic in ['mean', 'max', 'min', 'std', 'ptp']:
            stat_val = getattr(self.attributes[attribute][ti].ravel()[ma], statistic)()
        elif statistic == 'median':
            stat_val = np.median(self.attributes[attribute][ti].ravel()[ma])
        elif statistic == "skew":
            stat_val = np.mean(self.attributes[attribute][ti].ravel()[ma]) - \
                       np.median(self.attributes[attribute][ti].ravel()[ma])
        elif 'percentile' in statistic:
            per = int(statistic.split("_")[1])
            stat_val = np.percentile(self.attributes[attribute][ti].ravel()[ma], per)
        elif 'dt' in statistic:
            stat_name = statistic[:-3]
            if ti == 0:
                stat_val = 0
            else:
                stat_val = self.calc_attribute_statistic(attribute, stat_name, time) \
                    - self.calc_attribute_statistic(attribute, stat_name, time - 1)
        else:
            stat_val = np.nan
        return stat_val

    def calc_timestep_statistic(self, statistic, time):
        """
        Calculate statistics from the primary attribute of the StObject.

        Args:
            statistic: statistic being calculated
            time: Timestep being investigated

        Returns:
            Value of the statistic
        """
        ti = np.where(self.times == time)[0][0]
        ma = np.where(self.masks[ti].ravel() == 1)
        if statistic in ['mean', 'max', 'min', 'std', 'ptp']:
            stat_val = getattr(self.timesteps[ti].ravel()[ma], statistic)()
        elif statistic == 'median':
            stat_val = np.median(self.timesteps[ti].ravel()[ma])
        elif 'percentile' in statistic:
            per = int(statistic.split("_")[1])
            stat_val = np.percentile(self.timesteps[ti].ravel()[ma], per)
        elif 'dt' in statistic:
            stat_name = statistic[:-3]
            if ti == 0:
                stat_val = 0
            else:
                stat_val = self.calc_timestep_statistic(stat_name, time) -\
                    self.calc_timestep_statistic(stat_name, time - 1)
        else:
            stat_val = np.nan
        return stat_val

    def calc_shape_statistics(self, stat_names):
        """
        Calculate shape statistics using regionprops applied to the object mask.
        
        Args:
            stat_names: List of statistics to be extracted from those calculated by regionprops.

        Returns:
            Dictionary of shape statistics
        """
        stats = {}
        try:
            all_props = [regionprops(m) for m in self.masks]
        except TypeError:
            print(self.masks)
            exit()
        for stat in stat_names:
            stats[stat] = np.mean([p[0][stat] for p in all_props])
        return stats

    def calc_shape_step(self, stat_names, time):
        """
        Calculate shape statistics for a single time step

        Args:
            stat_names: List of shape statistics calculated from region props
            time: Time being investigated

        Returns:
            List of shape statistics

        """
        ti = np.where(self.times == time)[0][0]
        props = regionprops(self.masks[ti], self.timesteps[ti])[0]
        shape_stats = []
        for stat_name in stat_names:
            if "moments_hu" in stat_name:
                hu_index = int(stat_name.split("_")[-1])
                hu_name = "_".join(stat_name.split("_")[:-1])
                hu_val = np.log(props[hu_name][hu_index])
                if np.isnan(hu_val):
                    shape_stats.append(0)
                else:
                    shape_stats.append(hu_val)
            else:
                shape_stats.append(props[stat_name])
        return shape_stats

    def to_geojson(self, filename, proj, metadata=None):
        """
        Output the data in the STObject to a geoJSON file.

        Args:
            filename: Name of the file
            proj: PyProj object for converting the x and y coordinates back to latitude and longitue values.
            metadata: Metadata describing the object to be included in the top-level properties.
        """
        if metadata is None:
            metadata = {}
        json_obj = {"type": "FeatureCollection", "features": [], "properties": {}}
        json_obj['properties']['times'] = self.times.tolist()
        json_obj['properties']['dx'] = self.dx
        json_obj['properties']['step'] = self.step
        json_obj['properties']['u'] = self.u.tolist()
        json_obj['properties']['v'] = self.v.tolist()
        for k, v in metadata.items():
            json_obj['properties'][k] = v
        for t, time in enumerate(self.times):
            feature = {"type": "Feature",
                       "geometry": {"type": "Polygon"},
                       "properties": {}}
            boundary_coords = self.boundary_polygon(time)
            lonlat = np.vstack(proj(boundary_coords[0], boundary_coords[1], inverse=True))
            lonlat_list = lonlat.T.tolist()
            if len(lonlat_list) > 0:
                lonlat_list.append(lonlat_list[0])
            feature["geometry"]["coordinates"] = [lonlat_list]
            for attr in ["timesteps", "masks", "x", "y", "i", "j"]:
                feature["properties"][attr] = getattr(self, attr)[t].tolist()
            feature["properties"]["attributes"] = {}
            for attr_name, steps in self.attributes.items():
                feature["properties"]["attributes"][attr_name] = steps[t].tolist()
            json_obj['features'].append(feature)
        file_obj = open(filename, "w")
        json.dump(json_obj, file_obj, indent=1, sort_keys=True)
        file_obj.close()
        return

def read_geojson(filename):
    """
    Reads a geojson file containing an STObject and initializes a new STObject from the information in the file.

    Args:
        filename: Name of the geojson file

    Returns:
        an STObject
    """
    json_file = open(filename)
    data = json.load(json_file)
    json_file.close()
    times = data["properties"]["times"]
    main_data = dict(timesteps=[], masks=[], x=[], y=[], i=[], j=[])
    attribute_data = dict()
    for feature in data["features"]:
        for main_name in main_data.keys():
            main_data[main_name].append(np.array(feature["properties"][main_name]))
        for k, v in feature["properties"]["attributes"].items():
            if k not in attribute_data.keys():
                attribute_data[k] = [np.array(v)]
            else:
                attribute_data[k].append(np.array(v))
    kwargs = {}
    for kw in ["dx", "step", "u", "v"]:
        if kw in data["properties"].keys():
            kwargs[kw] = data["properties"][kw]
    sto = STObject(main_data["timesteps"], main_data["masks"], main_data["x"], main_data["y"],
                   main_data["i"], main_data["j"], times[0], times[-1], **kwargs)
    for k, v in attribute_data.items():
        sto.attributes[k] = v
    return sto




from __future__ import division
from .SSEFModelGrid import SSEFModelGrid
from .VSEModelGrid import VSEModelGrid
from .NCARModelGrid import NCARModelGrid
from .HRRRModelGrid import HRRRModelGrid
from .HREFv2ModelGrid import HREFv2ModelGrid
from .NCARStormEventModelGrid import NCARStormEventModelGrid
from hagelslag.util.make_proj_grids import make_proj_grids, read_arps_map_file, read_ncar_map_file, get_proj_obj
from hagelslag.util.derived_vars import relative_humidity_pressure_level, melting_layer_height
import numpy as np
from scipy.spatial import cKDTree
from scipy.ndimage import gaussian_filter
from pyproj import Proj

class ModelOutput(object):
    """
    Container for the model output values and spatial coordinate information.

    Attributes:
        ensemble_name (str): Name of the ensemble being loaded. Currently supports 'NCAR' and 'SSEF'.
        member_name (str): Ensemble member being loaded.
        run_date (datetime): Date of the initial timestep of the model run.
        variable (str): Variable being loaded.
        start_date (datetime.datetime): Date of the first timestep loaded.
        end_date (datetime.datetime): Date of the last timestep loaded.
        path (str): Path to model output
        single_step (bool): If true, each model timestep is in a separate file
            If false, all timesteps are together in the same file.

        map_file (str): path to data map file
    """
    def __init__(self, 
                 ensemble_name, 
                 member_name, 
                 run_date, 
                 variable, 
                 start_date, 
                 end_date,
                 path,
                 map_file,
                 sector_ind_path=None,
                 single_step=True):
        self.ensemble_name = ensemble_name
        self.member_name = member_name
        self.run_date = run_date
        self.variable = variable
        self.start_date = start_date
        self.end_date = end_date
        self.start_hour = int((self.start_date - self.run_date).total_seconds()) // 3600
        self.end_hour = int((self.end_date - self.run_date).total_seconds()) // 3600
        self.data = None
        self.valid_dates = None
        self.path = path
        self.map_file = map_file
        self.sector_ind_path=sector_ind_path
        self.lat = None
        self.lon = None
        self.x = None
        self.y = None
        self.i = None
        self.j = None
        self.proj = None
        self.dx = None
        self.units = ""
        self.single_step = single_step

    def load_data(self):
        """
        Load the specified variable from the ensemble files, then close the files.
        """
        if self.ensemble_name.upper() == "SSEF":
            if self.variable[0:2] == "rh":
                pressure_level = self.variable[2:]
                relh_vars = ["sph", "tmp"]
                relh_vals = {}
                for var in relh_vars:
                    mg = SSEFModelGrid(self.member_name,
                                       self.run_date,
                                       var + pressure_level,
                                       self.start_date,
                                       self.end_date,
                                       self.path,
                                       single_step=self.single_step)
                    relh_vals[var], units = mg.load_data()
                    mg.close()
                self.data = relative_humidity_pressure_level(relh_vals["tmp"],
                                                             relh_vals["sph"],
                                                             float(pressure_level) * 100)
                self.units = "%"
            elif self.variable == "melth":
                input_vars = ["hgtsfc", "hgt700", "hgt500", "tmp700", "tmp500"]
                input_vals = {}
                for var in input_vars:
                    mg = SSEFModelGrid(self.member_name,
                                       self.run_date,
                                       var,
                                       self.start_date,
                                       self.end_date,
                                       self.path,
                                       single_step=self.single_step)
                    input_vals[var], units = mg.load_data()
                    mg.close()
                self.data = melting_layer_height(input_vals["hgtsfc"],
                                                 input_vals["hgt700"],
                                                 input_vals["hgt500"],
                                                 input_vals["tmp700"],
                                                 input_vals["tmp500"])
                self.units = "m"
            else:
                mg = SSEFModelGrid(self.member_name,
                                   self.run_date,
                                   self.variable,
                                   self.start_date,
                                   self.end_date,
                                   self.path,
                                   single_step=self.single_step)
                self.data, self.units = mg.load_data()
                mg.close()
        elif self.ensemble_name.upper() == "NCAR":
            mg = NCARModelGrid(self.member_name,
                               self.run_date,
                               self.variable,
                               self.start_date,
                               self.end_date,
                               self.path,
                               single_step=self.single_step)
            self.data, self.units = mg.load_data()
            mg.close()
        elif self.ensemble_name.upper() == "HREFV2":
            proj_dict, grid_dict = read_ncar_map_file(self.map_file)
            mapping_data = make_proj_grids(proj_dict, grid_dict)            
           
            mg = HREFv2ModelGrid(self.member_name,
                               self.run_date,
                               self.variable,
                               self.start_date,
                               self.end_date,
                               self.path,
                               mapping_data,
                               self.sector_ind_path,
                               single_step=self.single_step)

            self.data, self.units = mg.load_data()

        elif self.ensemble_name.upper() == "VSE":
            mg = VSEModelGrid(self.member_name,
                               self.run_date,
                               self.variable,
                               self.start_date,
                               self.end_date,
                               self.path,
                               single_step=self.single_step)
            self.data, self.units = mg.load_data()
            mg.close()
        elif self.ensemble_name.upper() == "HRRR":
            mg = HRRRModelGrid(self.run_date,
                               self.variable,
                               self.start_date,
                               self.end_date,
                               self.path)
            self.data, self.units = mg.load_data()
            mg.close()
        elif self.ensemble_name.upper() == "NCARSTORM":
            mg = NCARStormEventModelGrid(self.run_date,
                                         self.variable,
                                         self.start_date,
                                         self.end_date,
                                         self.path)
            self.data, self.units = mg.load_data()
            mg.close()
        else:
            print(self.ensemble_name + " not supported.")

    def load_map_info(self, map_file):
        """
        Load map projection information and create latitude, longitude, x, y, i, and j grids for the projection.

        Args:
            map_file: File specifying the projection information.
        """
        if self.ensemble_name.upper() == "SSEF":
            proj_dict, grid_dict = read_arps_map_file(map_file)
            self.dx = int(grid_dict["dx"])
            mapping_data = make_proj_grids(proj_dict, grid_dict)
            for m, v in mapping_data.items():
                setattr(self, m, v)
            self.i, self.j = np.indices(self.lon.shape)
            self.proj = get_proj_obj(proj_dict)
        elif self.ensemble_name.upper() in ["NCAR", "NCARSTORM", "HRRR", "VSE", "HREFV2"]:
            proj_dict, grid_dict = read_ncar_map_file(map_file)
            if self.member_name[0:7] == "1km_pbl": # Don't just look at the first 3 characters. You have to differentiate '1km_pbl1' and '1km_on_3km_pbl1'
                grid_dict["dx"] = 1000
                grid_dict["dy"] = 1000
                grid_dict["sw_lon"] = 258.697
                grid_dict["sw_lat"] = 23.999
                grid_dict["ne_lon"] = 282.868269206236
                grid_dict["ne_lat"] = 36.4822338520542 

            self.dx = int(grid_dict["dx"])
            mapping_data = make_proj_grids(proj_dict, grid_dict)
            for m, v in mapping_data.items():
                setattr(self, m, v)
            self.i, self.j = np.indices(self.lon.shape)
            self.proj = get_proj_obj(proj_dict)


    def period_neighborhood_probability(self, radius, smoothing, threshold, stride, x=None, y=None, dx=None):
        """
        Calculate the neighborhood probability over the full period of the forecast

        Args:
            radius: circular radius from each point in km
            smoothing: standard deviation of Gaussian smoother in grid points
            threshold: intensity of exceedance
            stride: number of grid points to skip for reduced neighborhood grid

        Returns:
            neighborhood probabilities
        """
        if x is None:
            x = self.x
            y = self.y
            dx = self.dx
        neighbor_x = x[::stride, ::stride] / 1000.0
        neighbor_y = y[::stride, ::stride] / 1000.0
        neighbor_kd_tree = cKDTree(np.vstack((neighbor_x.ravel(), neighbor_y.ravel())).T)
        neighbor_prob = np.zeros((neighbor_x.shape[0], neighbor_x.shape[1]))
        period_max = self.data.max(axis=0)
        valid_i, valid_j = np.where(period_max >= threshold)
        print(self.variable, len(valid_i))
        if len(valid_i) > 0:
            var_kd_tree = cKDTree(np.vstack((x[valid_i, valid_j] / 1000.0, y[valid_i, valid_j] / 1000.0)).T)
            exceed_points = np.unique(np.concatenate(var_kd_tree.query_ball_tree(neighbor_kd_tree, radius))).astype(int)
            print("Exceed points", len(exceed_points))
            exceed_i, exceed_j = np.unravel_index(exceed_points, neighbor_x.shape)
            neighbor_prob[exceed_i, exceed_j] = 1
            if smoothing > 0:
                neighbor_prob = gaussian_filter(neighbor_prob, smoothing)
        return neighbor_prob




from skimage.draw import polygon
import numpy as np
import shapefile
from netCDF4 import Dataset
from hagelslag.util.make_proj_grids import read_arps_map_file, read_ncar_map_file, make_proj_grids
from pyproj import Proj
import argparse
import matplotlib.pyplot as plt


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--shape", help="Shape file")
    parser.add_argument("-m", "--map", help="Map projection file")
    parser.add_argument("-o", "--out", help="Output netCDF file")
    args = parser.parse_args()
    print("Loading map info")
    mapping_data, proj_dict, grid_dict = create_map_grid(args.map)
    print("Creating mask grid")
    mask_grid = create_mask_grid(args.shape, mapping_data, proj_dict, grid_dict)
    output_netcdf_file(args.out, mask_grid, proj_dict, grid_dict)
    #plt.figure(figsize=(10, 6))
    #plt.contourf(mapping_data["lon"], mapping_data["lat"], mask_grid, cmap="Reds", vmin=0, vmax=1)
    #plt.show()
    return


def create_map_grid(map_file):
    if map_file[-3:] == "map":
        proj_dict, grid_dict = read_arps_map_file(map_file)
    else:
        proj_dict, grid_dict = read_ncar_map_file(map_file)
    mapping_data = make_proj_grids(proj_dict, grid_dict)
    return mapping_data, proj_dict, grid_dict


def create_mask_grid(mask_shape_file, mapping_data, proj_dict, grid_dict):
    map_proj = Proj(**proj_dict)
    offset_x = mapping_data["x"][0, 0]
    offset_y = mapping_data["y"][0, 0]
    mask_grid = np.zeros(mapping_data["lon"].shape, dtype=int)
    sf = shapefile.Reader(mask_shape_file)
    s = 0
    for state_shape in sf.shapeRecords():
        print(s, state_shape.record[4]) #changed [5])
        part_start = list(state_shape.shape.parts)
        part_end = list(state_shape.shape.parts[1:]) + [len(state_shape.shape.points)]
        for p in range(len(part_start)):
            lon_lat_points = np.array(state_shape.shape.points[part_start[p]:part_end[p]]).T
            sx, sy = map_proj(lon_lat_points[0], lon_lat_points[1])
            si = (sx - offset_x) / grid_dict["dx"]
            sj = (sy - offset_y) / grid_dict["dx"]
            yy, xx = polygon(sj, si)
            valid = np.where((yy < mask_grid.shape[0]) & (xx < mask_grid.shape[1]))
            mask_grid[yy[valid], xx[valid]] = 1
        s += 1
    return mask_grid


def output_netcdf_file(filename, mask_grid, proj_dict, grid_dict):
    out_set = Dataset(filename, "w")
    out_set.createDimension("y", mask_grid.shape[0])
    out_set.createDimension("x", mask_grid.shape[1])
    out_set.set_auto_mask(True)
    var = out_set.createVariable("usa_mask", 'u1', ("y", "x"), zlib=True)
    var[:] = mask_grid
    var.long_name = "USA mask (1 if by land, 0 if by sea)"
    for k, v in proj_dict.items():
        setattr(out_set, k, v)
    for k, v in grid_dict.items():
        setattr(out_set, k, v)
    out_set.close()


if __name__ == "__main__":
    main()



from hagelslag.data.ModelOutput import ModelOutput
from hagelslag.data.MRMSGrid import MRMSGrid
from hagelslag.processing.EnhancedWatershedSegmenter import EnhancedWatershed, rescale_data
from hagelslag.processing.tracker import label_storm_objects, extract_storm_patches, track_storms
from .ObjectMatcher import ObjectMatcher, TrackMatcher, TrackStepMatcher
from scipy.ndimage import find_objects, gaussian_filter
from .STObject import STObject, read_geojson
import numpy as np
from scipy.interpolate import interp1d
from glob import glob
import pandas as pd
from datetime import timedelta
from scipy.stats import gamma
from netCDF4 import Dataset


class TrackProcessor(object):
    """
    TrackProcessor identifies local maxima in a convection-allowing model run and links them in time to form
    storm tracks. A similar procedure is applied to the observations, and the two sets of tracks are matched.
    Storm and environmental attributes are extracted from within the identified track areas.
    
    Args:
        run_date: Datetime model run was initialized
        start_date: Datetime for the beginning of storm extraction.
        end_date: Datetime for the ending of storm extraction.
        ensemble_name: Name of the ensemble being used.
        ensemble_member: name of the ensemble member being used.
        variable: model variable being used for extraction.
        model_path: path to the ensemble output.
        model_map_file: File containing model map projection information.
        model_watershed_params: tuple of parameters used for EnhancedWatershed
        object_matcher_params: tuple of parameters used for ObjectMatcher.
        track_matcher_params: tuple of parameters for TrackMatcher or TrackStepMatcher.
        size_filter: minimum size of model objects
        gaussian_window: number of grid points
        match_steps: If True, match individual steps in tracks instead of matching whole tracks
        mrms_path: Path to MRMS netCDF files
        mrms_variable: MRMS variable being used
        mrms_watershed_params: tuple of parameters for Enhanced Watershed applied to MESH data.
        single_step: Whether model timesteps are in separate files or aggregated into one file.
        mask_file: netCDF filename containing a mask of valid grid points on the model domain.
    """
    def __init__(self,
                 run_date,
                 start_date,
                 end_date,
                 ensemble_name,
                 ensemble_member,
                 variable,
                 model_path,
                 model_map_file,
                 model_watershed_params,
                 object_matcher_params,
                 track_matcher_params,
                 size_filter,
                 gaussian_window,
                 sector_ind_path,
                 match_steps=True,
                 mrms_path=None,
                 mrms_variable=None,
                 mrms_watershed_params=None,
                 single_step=True,
                 mask_file=None,
                 patch_radius=32):
        self.run_date = run_date
        self.start_date = start_date
        self.end_date = end_date
        self.start_hour = int((self.start_date - self.run_date).total_seconds()) / 3600
        self.end_hour = int((self.end_date - self.run_date).total_seconds()) / 3600
        self.hours = np.arange(int(self.start_hour), int(self.end_hour) + 1)
        self.ensemble_name = ensemble_name
        self.ensemble_member = ensemble_member
        self.variable = variable
        self.model_ew = EnhancedWatershed(*model_watershed_params)
        self.object_matcher = ObjectMatcher(*object_matcher_params)
        if match_steps:
            self.track_matcher = None
            self.track_step_matcher = TrackStepMatcher(*track_matcher_params)
        else:
            self.track_matcher = TrackMatcher(*track_matcher_params)
            self.track_step_matcher = None
        self.size_filter = size_filter
        self.gaussian_window = gaussian_window
        self.sector_ind_path = sector_ind_path
        self.model_path = model_path
        self.model_map_file = model_map_file
        self.mrms_path = mrms_path
        self.single_step = single_step
        self.model_grid = ModelOutput(self.ensemble_name, self.ensemble_member, self.run_date, self.variable,
                                      self.start_date, self.end_date, self.model_path, self.model_map_file,
                                      self.sector_ind_path,single_step=self.single_step)
        self.model_grid.load_map_info(self.model_map_file)
        if self.mrms_path is not None:
            self.mrms_variable = mrms_variable
            self.mrms_grid = MRMSGrid(self.start_date, self.end_date, self.mrms_variable, self.mrms_path)
            self.mrms_ew = EnhancedWatershed(*mrms_watershed_params)
        else:
            self.mrms_grid = None
            self.mrms_ew = None
        self.mask_file = mask_file
        self.mask = None
        if self.mask_file is not None:
            mask_data = Dataset(self.mask_file)
            self.mask = mask_data.variables["usa_mask"][:]
            mask_data.close()
        self.patch_radius = patch_radius
        return

    def find_model_patch_tracks(self):
        """
        Identify storms in gridded model output and extract uniform sized patches around the storm centers of mass.

        Returns:

        """
        self.model_grid.load_data()
        tracked_model_objects = []
        model_objects = []
        if self.model_grid.data is None:
            print("No model output found")
            return tracked_model_objects
        min_orig = self.model_ew.min_thresh
        max_orig = self.model_ew.max_thresh
        data_increment_orig = self.model_ew.data_increment
        self.model_ew.min_thresh = 0
        self.model_ew.data_increment = 1
        self.model_ew.max_thresh = 100
        for h, hour in enumerate(self.hours):
            # Identify storms at each time step and apply size filter
            print("Finding {0} objects for run {1} Hour: {2:02d}".format(self.ensemble_member,
                                                                         self.run_date.strftime("%Y%m%d%H"), hour))
            if self.mask is not None:
                model_data = self.model_grid.data[h] * self.mask
            else:
                model_data = self.model_grid.data[h]
            model_data[:self.patch_radius] = 0
            model_data[-self.patch_radius:] = 0
            model_data[:, :self.patch_radius] = 0
            model_data[:, -self.patch_radius:] = 0
            scaled_data = np.array(rescale_data(model_data, min_orig, max_orig))
            hour_labels = label_storm_objects(scaled_data, "ew",
                                              self.model_ew.min_thresh, self.model_ew.max_thresh,
                                              min_area=self.size_filter, max_area=self.model_ew.max_size,
                                              max_range=self.model_ew.delta, increment=self.model_ew.data_increment,
                                              gaussian_sd=self.gaussian_window)
            model_objects.extend(extract_storm_patches(hour_labels, model_data, self.model_grid.x,
                                                       self.model_grid.y, [hour],
                                                       dx=self.model_grid.dx,
                                                       patch_radius=self.patch_radius))
            for model_obj in model_objects[-1]:
                dims = model_obj.timesteps[-1].shape
                if h > 0:
                    model_obj.estimate_motion(hour, self.model_grid.data[h-1], dims[1], dims[0])
            del scaled_data
            del model_data
            del hour_labels
        tracked_model_objects.extend(track_storms(model_objects, self.hours,
                                                  self.object_matcher.cost_function_components,
                                                  self.object_matcher.max_values,
                                                  self.object_matcher.weights))
        self.model_ew.min_thresh = min_orig
        self.model_ew.max_thresh = max_orig
        self.model_ew.data_increment = data_increment_orig
        return tracked_model_objects

    def find_model_tracks(self):
        """
        Identify storms at each model time step and link them together with object matching.

        Returns:
            List of STObjects containing model track information.
        """
        self.model_grid.load_data()
        model_objects = []
        tracked_model_objects = []
        if self.model_grid.data is None:
            print("No model output found")
            return tracked_model_objects
        for h, hour in enumerate(self.hours):
            # Identify storms at each time step and apply size filter
            print("Finding {0} objects for run {1} Hour: {2:02d}".format(self.ensemble_member,
                                                                         self.run_date.strftime("%Y%m%d%H"), hour))
            if self.mask is not None:
                model_data = self.model_grid.data[h] * self.mask
            else:
                model_data = self.model_grid.data[h]

            # remember orig values
            min_orig = self.model_ew.min_thresh
            max_orig = self.model_ew.max_thresh
            data_increment_orig = self.model_ew.data_increment
            # scale to int 0-100.
            scaled_data = np.array(rescale_data( self.model_grid.data[h], min_orig, max_orig))
            self.model_ew.min_thresh = 0
            self.model_ew.data_increment = 1
            self.model_ew.max_thresh = 100
            hour_labels = self.model_ew.label(gaussian_filter(scaled_data, self.gaussian_window))
            hour_labels[model_data < self.model_ew.min_thresh] = 0
            hour_labels = self.model_ew.size_filter(hour_labels, self.size_filter)
            # Return to orig values
            self.model_ew.min_thresh = min_orig
            self.model_ew.max_thresh = max_orig
            self.model_ew.data_increment = data_increment_orig
            obj_slices = find_objects(hour_labels)

            num_slices = len(obj_slices)
            model_objects.append([])
            if num_slices > 0:
                for s, sl in enumerate(obj_slices):
                    model_objects[-1].append(STObject(self.model_grid.data[h][sl],
                                                      np.where(hour_labels[sl] == s + 1, 1, 0),
                                                      self.model_grid.x[sl], 
                                                      self.model_grid.y[sl], 
                                                      self.model_grid.i[sl], 
                                                      self.model_grid.j[sl],
                                                      hour,
                                                      hour,
                                                      dx=self.model_grid.dx))
                    if h > 0:
                        dims = model_objects[-1][-1].timesteps[0].shape
                        model_objects[-1][-1].estimate_motion(hour, self.model_grid.data[h-1], dims[1], dims[0])
            del hour_labels
            del scaled_data
            del model_data
        for h, hour in enumerate(self.hours):
            past_time_objs = []
            for obj in tracked_model_objects:
                # Potential trackable objects are identified
                if obj.end_time == hour - 1:
                    past_time_objs.append(obj)
            # If no objects existed in the last time step, then consider objects in current time step all new
            if len(past_time_objs) == 0:
                tracked_model_objects.extend(model_objects[h])
            # Match from previous time step with current time step
            elif len(past_time_objs) > 0 and len(model_objects[h]) > 0:
                assignments = self.object_matcher.match_objects(past_time_objs, model_objects[h], hour - 1, hour)
                unpaired = list(range(len(model_objects[h])))
                for pair in assignments:
                    past_time_objs[pair[0]].extend(model_objects[h][pair[1]])
                    unpaired.remove(pair[1])
                if len(unpaired) > 0:
                    for up in unpaired:
                        tracked_model_objects.append(model_objects[h][up])
            print("Tracked Model Objects: {0:03d} Hour: {1:02d}".format(len(tracked_model_objects), hour))

        return tracked_model_objects

    def load_model_tracks(self, json_path):
        model_track_files = sorted(glob(json_path + "{0}/{1}/{2}_*.json".format(self.run_date.strftime("%Y%m%d"),
                                                                                self.ensemble_member,
                                                                                self.ensemble_name)))
        model_tracks = []
        for model_track_file in model_track_files:
            model_tracks.append(read_geojson(model_track_file))
        return model_tracks

    def load_mrms_tracks(self, json_path, mrms_name="mesh"):
        mrms_track_files = sorted(glob(json_path + "{0}/{1}/{2}_*.json".format(self.run_date.strftime("%Y%m%d"),
                                                                               self.ensemble_member,
                                                                               mrms_name)))
        mrms_tracks = []
        for mrms_track_file in mrms_track_files:
            mrms_tracks.append(read_geojson(mrms_track_file))
        return mrms_tracks

    def find_mrms_tracks(self):
        """
        Identify objects from MRMS timesteps and link them together with object matching.

        Returns:
            List of STObjects containing MESH track information.
        """
        obs_objects = []
        tracked_obs_objects = []
        if self.mrms_ew is not None:
            self.mrms_grid.load_data()
            
            if len(self.mrms_grid.data) != len(self.hours):
                print('Less than 24 hours of observation data found')
                
                return tracked_obs_objects
         
            for h, hour in enumerate(self.hours):
                mrms_data = np.zeros(self.mrms_grid.data[h].shape)
                mrms_data[:] = np.array(self.mrms_grid.data[h])
                mrms_data[mrms_data < 0] = 0
                hour_labels = self.mrms_ew.size_filter(self.mrms_ew.label(gaussian_filter(mrms_data,
                                                                                      self.gaussian_window)),
                                                       self.size_filter)
                hour_labels[mrms_data < self.mrms_ew.min_thresh] = 0
                obj_slices = find_objects(hour_labels)
                num_slices = len(obj_slices)
                obs_objects.append([])
                if num_slices > 0:
                    for sl in obj_slices:
                        obs_objects[-1].append(STObject(mrms_data[sl],
                                                        np.where(hour_labels[sl] > 0, 1, 0),
                                                        self.model_grid.x[sl],
                                                        self.model_grid.y[sl],
                                                        self.model_grid.i[sl],
                                                        self.model_grid.j[sl],
                                                        hour,
                                                        hour,
                                                        dx=self.model_grid.dx))
                        if h > 0:
                            dims = obs_objects[-1][-1].timesteps[0].shape
                            obs_objects[-1][-1].estimate_motion(hour, self.mrms_grid.data[h-1], dims[1], dims[0])
        
            for h, hour in enumerate(self.hours):
                past_time_objs = []
                for obj in tracked_obs_objects:
                    if obj.end_time == hour - 1:
                        past_time_objs.append(obj)
                if len(past_time_objs) == 0:
                    tracked_obs_objects.extend(obs_objects[h])
                elif len(past_time_objs) > 0 and len(obs_objects[h]) > 0:
                    assignments = self.object_matcher.match_objects(past_time_objs, obs_objects[h], hour - 1, hour)
                    unpaired = list(range(len(obs_objects[h])))
                    for pair in assignments:
                        past_time_objs[pair[0]].extend(obs_objects[h][pair[1]])
                        unpaired.remove(pair[1])
                    if len(unpaired) > 0:
                        for up in unpaired:
                            tracked_obs_objects.append(obs_objects[h][up])
                print("Tracked Obs Objects: {0:03d} Hour: {1:02d}".format(len(tracked_obs_objects), hour))
        
        return tracked_obs_objects

    def match_tracks(self, model_tracks, obs_tracks, unique_matches=True, closest_matches=False):
        """
        Match forecast and observed tracks.

        Args:
            model_tracks:
            obs_tracks:
            unique_matches:
            closest_matches:

        Returns:

        """
        if unique_matches:
            pairings = self.track_matcher.match_tracks(model_tracks, obs_tracks, closest_matches=closest_matches)
        else:
            pairings = self.track_matcher.neighbor_matches(model_tracks, obs_tracks)
        return pairings

    def match_track_steps(self, model_tracks, obs_tracks):
        return self.track_step_matcher.match(model_tracks, obs_tracks)

    def extract_model_attributes(self, tracked_model_objects, storm_variables, potential_variables,
                                 tendency_variables=None, future_variables=None):
        """
        Extract model attribute data for each model track. Storm variables are those that describe the model storm
        directly, such as radar reflectivity or updraft helicity. Potential variables describe the surrounding
        environmental conditions of the storm, and should be extracted from the timestep before the storm arrives to
        reduce the chance of the storm contaminating the environmental values. Examples of potential variables include
        CAPE, shear, temperature, and dewpoint. Future variables are fields that occur in the hour after the extracted
        field.

        Args:
            tracked_model_objects: List of STObjects describing each forecasted storm
            storm_variables: List of storm variable names
            potential_variables: List of potential variable names.
            tendency_variables: List of tendency variables
        """
        if tendency_variables is None:
            tendency_variables = []
        if future_variables is None:
            future_variables = []
        model_grids = {}
        for l_var in ["lon", "lat"]:
            for model_obj in tracked_model_objects:
                model_obj.extract_attribute_array(getattr(self.model_grid, l_var), l_var)
        for storm_var in storm_variables:
            print("Storm {0} {1} {2}".format(storm_var,self.ensemble_member, self.run_date.strftime("%Y%m%d")))
            model_grids[storm_var] = ModelOutput(self.ensemble_name, self.ensemble_member,
                                                 self.run_date, storm_var, self.start_date - timedelta(hours=1),
                                                 self.end_date + timedelta(hours=1),
                                                 self.model_path,self.model_map_file, 
                                                 self.sector_ind_path,self.single_step)
            model_grids[storm_var].load_data()
            for model_obj in tracked_model_objects:
                model_obj.extract_attribute_grid(model_grids[storm_var])
            if storm_var not in potential_variables + tendency_variables + future_variables:
                del model_grids[storm_var]
        for potential_var in potential_variables:
            print("Potential {0} {1} {2}".format(potential_var,self.ensemble_member, self.run_date.strftime("%Y%m%d")))
            if potential_var not in model_grids.keys():
                model_grids[potential_var] = ModelOutput(self.ensemble_name, self.ensemble_member,
                                                         self.run_date, potential_var,
                                                         self.start_date - timedelta(hours=1),
                                                         self.end_date + timedelta(hours=1),
                                                         self.model_path, self.model_map_file, 
                                                         self.sector_ind_path,self.single_step)
                model_grids[potential_var].load_data()
            for model_obj in tracked_model_objects:
                model_obj.extract_attribute_grid(model_grids[potential_var], potential=True)
            if potential_var not in tendency_variables + future_variables:
                del model_grids[potential_var]
        for future_var in future_variables:
            print("Future {0} {1} {2}".format(future_var, self.ensemble_member, self.run_date.strftime("%Y%m%d")))
            if future_var not in model_grids.keys():
                model_grids[future_var] = ModelOutput(self.ensemble_name, self.ensemble_member,
                                                         self.run_date, future_var,
                                                         self.start_date - timedelta(hours=1),
                                                         self.end_date + timedelta(hours=1),
                                                         self.model_path, self.model_map_file,
                                                         self.sector_ind_path,self.single_step)
                model_grids[future_var].load_data()
            for model_obj in tracked_model_objects:
                model_obj.extract_attribute_grid(model_grids[future_var], future=True)
            if future_var not in tendency_variables:
                del model_grids[future_var]
        for tendency_var in tendency_variables:
            print("Tendency {0} {1} {2}".format(tendency_var, self.ensemble_member, self.run_date.strftime("%Y%m%d")))
            if tendency_var not in model_grids.keys():
                model_grids[tendency_var] = ModelOutput(self.ensemble_name, self.ensemble_member,
                                                        self.run_date, tendency_var,
                                                        self.start_date - timedelta(hours=1),
                                                        self.end_date,
                                                        self.model_path, self.model_map_file, 
                                                        self.sector_ind_path,self.single_step)
            for model_obj in tracked_model_objects:
                model_obj.extract_tendency_grid(model_grids[tendency_var])
            del model_grids[tendency_var]


    @staticmethod
    def match_hail_sizes(model_tracks, obs_tracks, track_pairings):
        """
        Given forecast and observed track pairings, maximum hail sizes are associated with each paired forecast storm
        track timestep. If the duration of the forecast and observed tracks differ, then interpolation is used for the
        intermediate timesteps.

        Args:
            model_tracks: List of model track STObjects
            obs_tracks: List of observed STObjects
            track_pairings: list of tuples containing the indices of the paired (forecast, observed) tracks
        """
        unpaired = list(range(len(model_tracks)))
        for p, pair in enumerate(track_pairings):
            model_track = model_tracks[pair[0]]
            unpaired.remove(pair[0])
            obs_track = obs_tracks[pair[1]]
            obs_hail_sizes = np.array([step[obs_track.masks[t] == 1].max()
                                       for t, step in enumerate(obs_track.timesteps)])
            if obs_track.times.size > 1 and model_track.times.size > 1:
                normalized_obs_times = 1.0 / (obs_track.times.max() - obs_track.times.min())\
                    * (obs_track.times - obs_track.times.min())
                normalized_model_times = 1.0 / (model_track.times.max() - model_track.times.min())\
                    * (model_track.times - model_track.times.min())
                hail_interp = interp1d(normalized_obs_times, obs_hail_sizes, kind="nearest",
                                       bounds_error=False, fill_value=0)
                model_track.observations = hail_interp(normalized_model_times)
            elif obs_track.times.size == 1:
                model_track.observations = np.ones(model_track.times.shape) * obs_hail_sizes[0]
            elif model_track.times.size == 1:
                model_track.observations = np.array([obs_hail_sizes.max()])
            print(pair[0], "obs",  obs_hail_sizes)
            print(pair[0], "model", model_track.observations)
        for u in unpaired:
            model_tracks[u].observations = np.zeros(model_tracks[u].times.shape)

    def match_size_distributions(self, model_tracks, obs_tracks, track_pairings):
        def match_single_track_dist(model_track, obs_track):
            label_columns = ["Max_Hail_Size", "Shape", "Location", "Scale"]
            obs_hail_dists = pd.DataFrame(index=obs_track.times,
                                          columns=label_columns)
            model_hail_dists = pd.DataFrame(index=model_track.times,
                                            columns=label_columns)
            for t, step in enumerate(obs_track.timesteps):
                step_vals = step[(obs_track.masks[t] == 1) & (obs_track.timesteps[t] > self.mrms_ew.min_thresh)]
                min_hail = step_vals.min() - 0.1
                obs_hail_dists.loc[obs_track.times[t], ["Shape", "Location", "Scale"]] = gamma.fit(step_vals,
                                                                                                   floc=min_hail)
                obs_hail_dists.loc[obs_track.times[t], "Max_Hail_Size"] = step_vals.max()
            if obs_track.times.size > 1 and model_track.times.size > 1:
                normalized_obs_times = 1.0 / (obs_track.times.max() - obs_track.times.min()) \
                                       * (obs_track.times - obs_track.times.min())
                normalized_model_times = 1.0 / (model_track.times.max() - model_track.times.min()) \
                                         * (model_track.times - model_track.times.min())
                for col in label_columns:
                    interp_func = interp1d(normalized_obs_times, obs_hail_dists[col], kind="linear",
                                           bounds_error=False, fill_value=0)
                    model_hail_dists.loc[model_track.times, col] = interp_func(normalized_model_times)
            else:
                for param in obs_hail_dists.columns:
                    model_hail_dists.loc[model_track.times, param] = obs_hail_dists.loc[obs_track.times[0], param]
            return model_hail_dists
        unpaired = list(range(len(model_tracks)))
        for p, pair in enumerate(track_pairings):
            unpaired.remove(pair[0])
            if type(pair[1]) in [int, np.int64, np.int32]:
                interp_hail_dists = match_single_track_dist(model_tracks[pair[0]], obs_tracks[pair[1]])
                model_tracks[pair[0]].observations = interp_hail_dists
            else:
                model_tracks[pair[0]].observations = []
                for op in pair[1]:
                    interp_hail_dists = match_single_track_dist(model_tracks[pair[0]], obs_tracks[op])
                    model_tracks[pair[0]].observations.append(interp_hail_dists)
        return

    def match_hail_size_step_distributions(self, model_tracks, obs_tracks, track_pairings):
        """
        Given a matching set of observed tracks for each model track, 
        
        Args:
            model_tracks: 
            obs_tracks: 
            track_pairings: 

        Returns:

        """
        label_columns = ["Matched", "Max_Hail_Size", "Num_Matches", "Shape", "Location", "Scale"]
        s = 0
        for m, model_track in enumerate(model_tracks):
            model_track.observations = pd.DataFrame(index=model_track.times, columns=label_columns, dtype=np.float64)
            model_track.observations.loc[:, :] = 0
            model_track.observations["Matched"] = model_track.observations["Matched"].astype(np.int32)
            for t, time in enumerate(model_track.times):
                model_track.observations.loc[time, "Matched"] = track_pairings.loc[s, "Matched"]
                if model_track.observations.loc[time, "Matched"] > 0:
                    all_hail_sizes = []
                    step_pairs = track_pairings.loc[s, "Pairings"]
                    for step_pair in step_pairs:
                        obs_step = obs_tracks[step_pair[0]].timesteps[step_pair[1]].ravel()
                        obs_mask = obs_tracks[step_pair[0]].masks[step_pair[1]].ravel()
                        all_hail_sizes.append(obs_step[(obs_mask == 1) & (obs_step >= self.mrms_ew.min_thresh)])
                    combined_hail_sizes = np.concatenate(all_hail_sizes)
                    min_hail = combined_hail_sizes.min() - 0.1
                    model_track.observations.loc[time, "Max_Hail_Size"] = combined_hail_sizes.max()
                    model_track.observations.loc[time, "Num_Matches"] = step_pairs.shape[0]
                    model_track.observations.loc[time, ["Shape", "Location", "Scale"]] = gamma.fit(combined_hail_sizes,
                                                                                                   floc=min_hail)
                s += 1

    @staticmethod
    def calc_track_errors(model_tracks, obs_tracks, track_pairings):
        """
        Calculates spatial and temporal translation errors between matched
        forecast and observed tracks.

        Args:
            model_tracks: List of model track STObjects
            obs_tracks: List of observed track STObjects
            track_pairings: List of tuples pairing forecast and observed tracks.

        Returns:
            pandas DataFrame containing different track errors
        """
        columns = ['obs_track_id',
                   'translation_error_x',
                   'translation_error_y',
                   'start_time_difference',
                   'end_time_difference',
                   ]
        track_errors = pd.DataFrame(index=list(range(len(model_tracks))),
                                    columns=columns)
        for p, pair in enumerate(track_pairings):
            model_track = model_tracks[pair[0]]
            if type(pair[1]) in [int, np.int64]:
                obs_track = obs_tracks[pair[1]]
            else:
                obs_track = obs_tracks[pair[1][0]]
            model_com = model_track.center_of_mass(model_track.start_time)
            obs_com = obs_track.center_of_mass(obs_track.start_time)
            track_errors.loc[pair[0], 'obs_track_id'] = pair[1] if type(pair[1]) in [int, np.int64] else pair[1][0]
            track_errors.loc[pair[0], 'translation_error_x'] = model_com[0] - obs_com[0]
            track_errors.loc[pair[0], 'translation_error_y'] = model_com[1] - obs_com[1]
            track_errors.loc[pair[0], 'start_time_difference'] = model_track.start_time - obs_track.start_time
            track_errors.loc[pair[0], 'end_time_difference'] = model_track.end_time - obs_track.end_time 
        return track_errors




from .ModelGrid import ModelGrid
import numpy as np
from datetime import timedelta


class NCARModelGrid(ModelGrid):
    """
    Extension of the ModelGrid class for interfacing with the NCAR ensemble.

    Args:
        member (str): Name of the ensemble member
        run_date (datetime.datetime object): Date of the initial step of the ensemble run
        start_date (datetime.datetime object): First time step extracted.
        end_date (datetime.datetime object): Last time step extracted.
        path (str): Path to model output files.
        single_step (boolean (default=False)): Whether variable information is stored with each time step in a separate
            file or one file containing all timesteps.
    """
    def __init__(self, member, run_date, variable, start_date, end_date, path, single_step=False):
        self.member = member
        self.path = path
        self.forecast_hours = np.arange((start_date - run_date).total_seconds() / 3600,
                                        (end_date - run_date).total_seconds() / 3600 + 1, dtype=int)
        filenames = []
        if not single_step:
            filenames.append("{0}{1}/{2}_surrogate_{1}.nc".format(self.path,
                                                                  run_date.strftime("%Y%m%d%H"),
                                                                  self.member))
            load_var = variable
        else:
            for hour in self.forecast_hours:
                valid_time = run_date + timedelta(hours=hour)
                if variable in ["UBSHR1", "VBSHR1", "UBSHR6", "VBSHR6", "PWAT", "SRH3", "LCL_HEIGHT", "CAPE_SFC",
                                "CIN_SFC", "MUCAPE"]:
                    filenames.append("{0}{1}/post_rundir/{2}/fhr_{3:d}/WRFTWO{3:02d}.nc".format(self.path,
                                                                                                run_date.strftime(
                                                                                                    "%Y%m%d%H"),
                                                                                                self.member.replace(
                                                                                                    "mem", "mem_"),
                                                                                                int(hour)))
                else:
                    filenames.append("{0}{1}/wrf_rundir/{2}/wrfout_d02_{3}:00:00".format(self.path,
                                                                                         run_date.strftime("%Y%m%d%H"),
                                                                                         self.member.replace("mem",
                                                                                                             "ens_"),
                                                                                         valid_time.strftime(
                                                                                             "%Y-%m-%d_%H")))
            if variable == "SRH3":
                load_var = "SR_HELICITY_3KM"
            elif variable == "CAPE_SFC":
                load_var = "SBCAPE"
            elif variable == "CIN_SFC":
                load_var = "SBCINH"
            else:
                load_var = variable
        super(NCARModelGrid, self).__init__(filenames, run_date, start_date, end_date, load_var)



import unittest
from hagelslag.data.MRMSGrid import MRMSGrid
from datetime import datetime

class TestMRMSGrid(unittest.TestCase):
    def setUp(self):
        self.path = "../../testdata/mrms_3km/"
        self.variable = "MESH_Max_60min_00.50"
        self.start_date = datetime(2015, 5, 1, 18, 0)
        self.end_date = datetime(2015, 5, 2, 15, 0)
        self.mrms = MRMSGrid(self.start_date, self.end_date, self.variable, self.path)
 
    def test_constructor(self):
        self.assertEquals(self.mrms.all_dates.size, 22, "Number of dates is wrong")
        self.assertIsNone(self.mrms.data, "Data already loaded")         
        self.assertIsNone(self.mrms.valid_dates, "Valid dates already loaded")

    #def test_loading(self):
        #self.mrms.load_data()
        #self.assertEquals(self.mrms.data.shape[0], self.mrms.valid_dates.shape[0], "Data and valid dates unequal length")
        #self.assertEquals(self.mrms.all_dates.shape[0], self.mrms.valid_dates.shape[0], "All dates were not loaded")

        



import Nio
import pandas as pd
import os
import subprocess
import numpy as np
from scipy.interpolate import RectBivariateSpline
from scipy.spatial import cKDTree
import pickle
from netCDF4 import Dataset, date2num
from datetime import datetime, timedelta
from hagelslag.util.make_proj_grids import read_arps_map_file, read_ncar_map_file, make_proj_grids
from multiprocessing import Pool
import warnings
import traceback
import argparse


def main():
    warnings.simplefilter("ignore")
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--start", required=True, help="Date to begin aggregation in YYYYMMDDHHMM format")
    parser.add_argument("-e", "--end", required=True, help="Date to end aggregation in YYYYMMDDHHMM format")
    parser.add_argument("-p", "--path", required=True, help="Path to MRMS GRIB2 files")
    parser.add_argument("-o", "--out", required=True, help="Path to save MRMS netCDF4 files")
    parser.add_argument("-m", "--map", required=True, help="File containing map coordinates for interpolation")
    parser.add_argument("-v", "--var", required=True,
                        help="Comma-separated list of variables (example: MESH_Max_60min_00.50)")
    parser.add_argument("-i", "--int", default="max", help="Interpolation type ('max' or 'spline')")
    parser.add_argument("-n", "--np", type=int, default=1, help="Number of processors to use")
    args = parser.parse_args()
    start_date = datetime.strptime(args.start, "%Y%m%d%H%M")
    end_date = datetime.strptime(args.end, "%Y%m%d%H%M")
    mrms_path = args.path
    out_path = args.out
    map_filename = args.map
    variables = args.var.split(",")
    num_procs = args.np
    if num_procs > 1:
        pool = Pool(num_procs)
        curr_date = start_date
        while curr_date <= end_date:
            for variable in variables:
                pool.apply_async(interpolate_mrms_day, (curr_date, variable, args.int, mrms_path,
                                                        map_filename, out_path))
            curr_date += timedelta(days=1)
        pool.close()
        pool.join()
    else:
        curr_date = start_date
        while curr_date <= end_date:
            for variable in variables:
                interpolate_mrms_day(curr_date, variable, args.int, mrms_path, map_filename, out_path)
            curr_date += timedelta(days=1)


def load_map_coordinates(map_file):
    """
    Loads map coordinates from netCDF or pickle file created by util.makeMapGrids.

    Args:
        map_file: Filename for the file containing coordinate information.

    Returns:
        Latitude and longitude grids as numpy arrays.
    """
    if map_file[-4:] == ".pkl":
        map_data = pickle.load(open(map_file))
        lon = map_data['lon']
        lat = map_data['lat']
    else:
        map_data = Dataset(map_file)
        if "lon" in map_data.variables.keys():
            lon = map_data.variables['lon'][:]
            lat = map_data.variables['lat'][:]
        else:
            lon = map_data.variables["XLONG"][0]
            lat = map_data.variables["XLAT"][0]
    return lon, lat


def interpolate_mrms_day(start_date, variable, interp_type, mrms_path, map_filename, out_path):
    """
    For a given day, this module interpolates hourly MRMS data to a specified latitude and 
    longitude grid, and saves the interpolated grids to CF-compliant netCDF4 files.
    
    Args:
        start_date (datetime.datetime): Date of data being interpolated
        variable (str): MRMS variable
        interp_type (str): Whether to use maximum neighbor or spline
        mrms_path (str): Path to top-level directory of MRMS GRIB2 files
        map_filename (str): Name of the map filename. Supports ARPS map file format and netCDF files containing latitude
            and longitude variables
        out_path (str): Path to location where interpolated netCDF4 files are saved.
    """
    try:
        print(start_date, variable)
        end_date = start_date + timedelta(hours=23)
        mrms = MRMSGrid(start_date, end_date, variable, mrms_path)
        if mrms.data is not None:
            if map_filename[-3:] == "map":
                mapping_data = make_proj_grids(*read_arps_map_file(map_filename))
                mrms.interpolate_to_netcdf(mapping_data['lon'], mapping_data['lat'], out_path, interp_type=interp_type)
            elif map_filename[-3:] == "txt":
                mapping_data = make_proj_grids(*read_ncar_map_file(map_filename))
                mrms.interpolate_to_netcdf(mapping_data["lon"], mapping_data["lat"], out_path, interp_type=interp_type)
            else:
                lon, lat = load_map_coordinates(map_filename)
                mrms.interpolate_to_netcdf(lon, lat, out_path, interp_type=interp_type)
    except Exception as e:
        # This exception catches any errors when run in multiprocessing, prints the stack trace,
        # and ends the process. Otherwise the process will stall.
        print(traceback.format_exc())
        raise e


class MRMSGrid(object):
    """
    MRMSGrid reads time series of MRMS grib2 files, interpolates them, and outputs them to netCDF4 format.

    """
    def __init__(self, start_date, end_date, variable, path_start, freq="1H"):
        self.start_date = start_date
        self.end_date = end_date
        self.variable = variable
        self.path_start = path_start
        self.freq = freq
        self.data = None
        self.all_dates = pd.DatetimeIndex(start=self.start_date, end=self.end_date, freq=self.freq)
        self.loaded_dates = None
        self.lon = None
        self.lat = None
        self.load_data()

    def load_data(self):
        """
        Loads data from MRMS GRIB2 files and handles compression duties if files are compressed.
        """
        data = []
        loaded_dates = []
        loaded_indices = []
        for t, timestamp in enumerate(self.all_dates):
            date_str = timestamp.date().strftime("%Y%m%d")
            full_path = self.path_start + date_str + "/"
            if self.variable in os.listdir(full_path):
                full_path += self.variable + "/"
                data_files = sorted(os.listdir(full_path))
                file_dates = pd.to_datetime([d.split("_")[-1][0:13] for d in data_files])
                if timestamp in file_dates:
                    data_file = data_files[np.where(timestamp==file_dates)[0][0]]
                    print(full_path + data_file)
                    if data_file[-2:] == "gz":
                        subprocess.call(["gunzip", full_path + data_file])
                        file_obj = Nio.open_file(full_path + data_file[:-3])
                    else:
                        file_obj = Nio.open_file(full_path + data_file)
                    var_name = sorted(file_obj.variables.keys())[0]
                    data.append(file_obj.variables[var_name][:])
                    if self.lon is None:
                        self.lon = file_obj.variables["lon_0"][:]
                        # Translates longitude values from 0:360 to -180:180
                        if np.count_nonzero(self.lon > 180) > 0:
                            self.lon -= 360
                        self.lat = file_obj.variables["lat_0"][:]
                    file_obj.close()
                    if data_file[-2:] == "gz":
                        subprocess.call(["gzip", full_path + data_file[:-3]])
                    else:
                        subprocess.call(["gzip", full_path + data_file])
                    loaded_dates.append(timestamp)
                    loaded_indices.append(t)
        if len(loaded_dates) > 0:
            self.loaded_dates = pd.DatetimeIndex(loaded_dates)
            self.data = np.ones((self.all_dates.shape[0], data[0].shape[0], data[0].shape[1])) * -9999
            self.data[loaded_indices] = np.array(data)

    def interpolate_grid(self, in_lon, in_lat):
        """
        Interpolates MRMS data to a different grid using cubic bivariate splines
        """
        out_data = np.zeros((self.data.shape[0], in_lon.shape[0], in_lon.shape[1]))
        for d in range(self.data.shape[0]):
            print("Loading ", d, self.variable, self.start_date)
            if self.data[d].max() > -999:
                step = self.data[d]
                step[step < 0] = 0
                if self.lat[-1] < self.lat[0]:
                    spline = RectBivariateSpline(self.lat[::-1], self.lon, step[::-1], kx=3, ky=3)
                else:
                    spline = RectBivariateSpline(self.lat, self.lon, step, kx=3, ky=3)
                print("Evaluating", d, self.variable, self.start_date)
                flat_data = spline.ev(in_lat.ravel(), in_lon.ravel())
                out_data[d] = flat_data.reshape(in_lon.shape)
                del spline
            else:
                print(d, " is missing")
                out_data[d] = -9999
        return out_data

    def max_neighbor(self, in_lon, in_lat, radius=0.05):
        """
        Finds the largest value within a given radius of a point on the interpolated grid.

        Args:
            in_lon: 2D array of longitude values
            in_lat: 2D array of latitude values
            radius: radius of influence for largest neighbor search in degrees

        Returns:
            Array of interpolated data
        """
        out_data = np.zeros((self.data.shape[0], in_lon.shape[0], in_lon.shape[1]))
        in_tree = cKDTree(np.vstack((in_lat.ravel(), in_lon.ravel())).T)
        out_indices = np.indices(out_data.shape[1:])
        out_rows = out_indices[0].ravel()
        out_cols = out_indices[1].ravel()
        for d in range(self.data.shape[0]):
            nz_points = np.where(self.data[d] > 0)
            if len(nz_points[0]) > 0:
                nz_vals = self.data[d][nz_points]
                nz_rank = np.argsort(nz_vals)
                original_points = cKDTree(np.vstack((self.lat[nz_points[0][nz_rank]], self.lon[nz_points[1][nz_rank]])).T)
                all_neighbors = original_points.query_ball_tree(in_tree, radius, p=2, eps=0)
                for n, neighbors in enumerate(all_neighbors):
                    if len(neighbors) > 0:
                        out_data[d, out_rows[neighbors], out_cols[neighbors]] = nz_vals[nz_rank][n]
        return out_data

    def interpolate_to_netcdf(self, in_lon, in_lat, out_path, date_unit="seconds since 1970-01-01T00:00",
                              interp_type="spline"):
        """
        Calls the interpolation function and then saves the MRMS data to a netCDF file. It will also create 
        separate directories for each variable if they are not already available.
        """
        if interp_type == "spline":
            out_data = self.interpolate_grid(in_lon, in_lat)
        else:
            out_data = self.max_neighbor(in_lon, in_lat)
        if not os.access(out_path + self.variable, os.R_OK):
            try:
                os.mkdir(out_path + self.variable)
            except OSError:
                print(out_path + self.variable + " already created")
        out_file = out_path + self.variable + "/" + "{0}_{1}_{2}.nc".format(self.variable,
                                                                            self.start_date.strftime("%Y%m%d-%H:%M"),
                                                                            self.end_date.strftime("%Y%m%d-%H:%M"))
        out_obj = Dataset(out_file, "w")
        out_obj.createDimension("time", out_data.shape[0])
        out_obj.createDimension("y", out_data.shape[1])
        out_obj.createDimension("x", out_data.shape[2])
        data_var = out_obj.createVariable(self.variable, "f4", ("time", "y", "x"), zlib=True, 
                                          fill_value=-9999.0,
                                          least_significant_digit=3)
        data_var[:] = out_data
        data_var.long_name = self.variable
        data_var.coordinates = "latitude longitude"
        if "MESH" in self.variable or "QPE" in self.variable:
            data_var.units = "mm"
        elif "Reflectivity" in self.variable:
            data_var.units = "dBZ"
        elif "Rotation" in self.variable:
            data_var.units = "s-1"
        else:
            data_var.units = ""
        out_lon = out_obj.createVariable("longitude", "f4", ("y", "x"), zlib=True)
        out_lon[:] = in_lon
        out_lon.units = "degrees_east"
        out_lat = out_obj.createVariable("latitude", "f4", ("y", "x"), zlib=True)
        out_lat[:] = in_lat
        out_lat.units = "degrees_north"
        dates = out_obj.createVariable("time", "i8", ("time",), zlib=True)
        dates[:] = np.round(date2num(self.all_dates.to_pydatetime(), date_unit)).astype(np.int64)
        dates.long_name = "Valid date"
        dates.units = date_unit
        out_obj.Conventions="CF-1.6"
        out_obj.close()
        return


if __name__ == "__main__":
    main()



import pickle
import numpy as np
from hagelslag.util.Config import Config
import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("config", help="Config file containing input columns")
    parser.add_argument("-m", "--model", required=True, help="List of machine learning models with feature importances")
    parser.add_argument("-n", "--num", type=int, default=10, help="Number of features to display")
    args = parser.parse_args()
    config = Config(args.config)
    input_columns = np.array(config.size_distribution_input_columns)
    model_files = args.model.split(",")
    for model_file in model_files:
        print(model_file)
        model_fo = open(model_file)
        model_obj = pickle.load(model_fo)
        model_fo.close()
        importances = None
        if hasattr(model_obj, "feature_importances_"):
            importances = model_obj.feature_importances_
        elif hasattr(model_obj, "best_estimator_"):
            if hasattr(model_obj.best_estimator_, "feature_importances_"):
                importances = model_obj.best_estimator_.feature_importances_
        if importances is not None:
            ranking = np.argsort(importances)[::-1]
            for r, rank in enumerate(ranking[:args.num]):
                print("{0:2d} {1} {2:1.3f}".format(r, input_columns[rank], importances[rank]))

    return


if __name__ == "__main__":
    main()



from netCDF4 import Dataset, num2date
import pandas as pd
import numpy as np
import os
from scipy.spatial import cKDTree
from scipy.ndimage import gaussian_filter


class MRMSGrid(object):
    """
    An interface to the NOAA National Severe Storms Lab Multi-Radar Multi-Sensor (MRMS) dataset.

    MRMSGrid assumes that the data are in netCDF format and have been interpolated to match the grid being used for
    forecasting.

    Args:
        start_date (datetime.datetime or time str): Date of first time step to be loaded.
        end_date (datetime.datetime or str in timestamp format): Date of last time step to be loaded.
        variable (str): MRMS variable name
        path (str): Path to the directory containing MRMS files.
        freq (str, optional (default="1H")): Time frequency of the data being loaded. Uses pandas time syntax.

    Attributes:
        start_date (datetime.datetime or time str): Date of first time step to be loaded.
        end_date (datetime.datetime or str in timestamp format): Date of last time step to be loaded.
        variable (str): MRMS variable name
        path (str): Path to the directory containing MRMS files.
        freq (str, optional (default="1H")): Time frequency of the data being loaded. Uses pandas time syntax.
        all_dates : pandas.DatetimeIndex
        List of dates being loaded
        data (ndarray or None): Array of gridded observations after load_data is called. None otherwise.
        valid_dates (ndarray): Contains the dates where data loaded successfully.
    """
    def __init__(self, start_date, end_date, variable, path, freq="1H"):
        self.start_date = start_date
        self.end_date = end_date
        self.variable = variable
        self.path = path
        self.freq = freq
        self.all_dates = pd.DatetimeIndex(start=start_date, end=end_date, freq=freq)
        self.data = None
        self.valid_dates = None

    def load_data(self):
        """
        Loads data files and stores the output in the data attribute.
        """
        data = []
        valid_dates = []
        mrms_files = np.array(sorted(os.listdir(self.path + self.variable + "/")))
        mrms_file_dates = np.array([m_file.split("_")[-2].split("-")[0]
            for m_file in mrms_files])
        old_mrms_file = None
        file_obj = None
        for t in range(self.all_dates.shape[0]):
            file_index = np.where(mrms_file_dates == self.all_dates[t].strftime("%Y%m%d"))[0]
            if len(file_index) > 0:
                mrms_file = mrms_files[file_index][0]
                if mrms_file is not None:
                    if file_obj is not None:
                        file_obj.close()
                    file_obj = Dataset(self.path + self.variable + "/" + mrms_file)
                    #old_mrms_file = mrms_file
                    
                    if "time" in file_obj.variables.keys():
                        time_var = "time"
                    else:
                        time_var = "date"
                    file_valid_dates = pd.DatetimeIndex(num2date(file_obj.variables[time_var][:],
                                                                 file_obj.variables[time_var].units))
                else:
                    file_valid_dates = pd.DatetimeIndex([])
                time_index = np.where(file_valid_dates.values == self.all_dates.values[t])[0]
                if len(time_index) > 0:
                    data.append(file_obj.variables[self.variable][time_index[0]])
                    valid_dates.append(self.all_dates[t])
        if file_obj is not None:
            file_obj.close()
        self.data = np.array(data)
        self.data[self.data < 0] = 0
        self.data[self.data > 150] = 150
        self.valid_dates = pd.DatetimeIndex(valid_dates)

    def period_neighborhood_probability(self, radius, smoothing, threshold, stride, x, y, dx):
        """
        Calculate the neighborhood probability over the full period of the forecast

        Args:
            radius: circular radius from each point in km
            smoothing: width of Gaussian smoother in km
            threshold: intensity of exceedance
            stride: number of grid points to skip for reduced neighborhood grid
            x: x-coordinate array in km
            y: y-coordinate array in km
            dx: distance between grid points in km

        Returns:
            neighborhood probablities
        """
        neighbor_x = x[::stride, ::stride]
        neighbor_y = y[::stride, ::stride]
        neighbor_kd_tree = cKDTree(np.vstack((neighbor_x.ravel(), neighbor_y.ravel())).T)
        neighbor_prob = np.zeros((neighbor_x.shape[0], neighbor_x.shape[1]))
        period_max = self.data.max(axis=0)
        valid_i, valid_j = np.where(period_max >= threshold)
        if len(valid_i) > 0:
            var_kd_tree = cKDTree(np.vstack((x[valid_i, valid_j], y[valid_i, valid_j])).T)
            exceed_points = np.unique(np.concatenate(var_kd_tree.query_ball_tree(neighbor_kd_tree, radius))).astype(int)
            exceed_i, exceed_j = np.unravel_index(exceed_points, neighbor_x.shape)
            neighbor_prob[exceed_i, exceed_j] = 1
            if smoothing > 0:
                neighbor_prob = gaussian_filter(neighbor_prob, int(smoothing / dx / stride))
        return neighbor_prob



import numpy as np
from scipy.ndimage import label, maximum, find_objects


class Hysteresis(object):
    """
    Object segmentation method that identifies objects as contiguous areas with all pixels above a low
    threshold and contain at least one pixel above a high threshold.

    Attributes:
        low_thresh: lower threshold value
        high_thresh: higher threshold value
    """

    def __init__(self, low_thresh, high_thresh):
        self.low_thresh = low_thresh
        self.high_thresh = high_thresh
        return

    def label(self, input_grid):
        """
        Label input grid with hysteresis method.

        Args:
            input_grid: 2D array of values.

        Returns:
            Labeled output grid.
        """
        unset = 0
        high_labels, num_labels = label(input_grid > self.high_thresh)
        region_ranking = np.argsort(maximum(input_grid, high_labels, index=np.arange(1, num_labels + 1)))[::-1]
        output_grid = np.zeros(input_grid.shape, dtype=int)
        stack = []
        for rank in region_ranking:
            label_num = rank + 1
            label_i, label_j = np.where(high_labels == label_num)
            for i in range(label_i.size):
                if output_grid[label_i[i], label_j[i]] == unset:
                    stack.append((label_i[i], label_j[i]))
            while len(stack) > 0:
                index = stack.pop()
                output_grid[index] = label_num
                for i in range(index[0] - 1, index[0] + 2):
                    for j in range(index[1] - 1, index[1] + 2):
                        if 0 <= i < output_grid.shape[0] and 0 <= j < output_grid.shape[1]:
                            if (input_grid[i, j] > self.low_thresh) and (output_grid[i, j] == unset):
                                stack.append((i, j))
        return output_grid

    @staticmethod
    def size_filter(labeled_grid, min_size):
        """
        Remove labeled objects that do not meet size threshold criteria.

        Args:
            labeled_grid: 2D output from label method.
            min_size: minimum size of object in pixels.

        Returns:
            labeled grid with smaller objects removed.
        """
        out_grid = np.zeros(labeled_grid.shape, dtype=int)
        slices = find_objects(labeled_grid)
        j = 1
        for i, s in enumerate(slices):
            box = labeled_grid[s]
            size = np.count_nonzero(box.ravel() == (i + 1))
            if size >= min_size and box.shape[0] > 1 and box.shape[1] > 1:
                out_grid[np.where(labeled_grid == i + 1)] = j
                j += 1
        return out_grid
